from __future__ import absolute_import

__description__ = "a module that houses object-based wrappers of keras based classifiers"
__author__ = "Kyle Rose (rosek@kenyon.edu)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import numpy as np

from idq import io
from idq import features

from . import base as base_classifiers

### required for keras functionality
from keras.models import Sequential as keras_sequential
from keras.models import load_model
from keras.models import save_model
from keras.layers import Dense as keras_dense
from keras.layers import Activation as keras_activation
from keras.models import Model as keras_Model
from keras.callbacks import Callback as keras_callback
from keras.callbacks import EarlyStopping as keras_earlystopping
from keras_pickle_wrapper import KerasPickleWrapper

#-------------------------------------------------

DEFAULT_COLUMN_VALUE = 0

#-------------------------------------------------
# base classes for keras classifiers
#-------------------------------------------------

class KerasModel(base_classifiers.ClassifierModel):
    """the model object for SupervisedKerasClassifier
    """

    def __init__(self, start, end, keras, whitener, channels, select, time=features.DEFAULT_TIME_NAME, segs=None):
        base_classifiers.ClassifierModel.__init__(self, start, end, segs=segs)

        ### define stuff particular to Keras
        self._keras = keras
        self._whitener = whitener
        self._channels = channels
        self._select = select
        self._time = time

    @property
    def keras(self):
        return self._keras

    @property
    def whitener(self):
        return self._whitener

    @property
    def channels(self):
        return self._channels

    @property
    def select(self):
        return self._select

    @propery
    def time(self):
        return self._time

#-------------------------------------------------

class SupervisedKerasClassifier(base_classifiers.SupervisedClassifier):
    """
    A base class for supervised keras classifiers
    Note: not a standalone classifier
    """
    _flavor = "keras_supervised_classifier"
    _required_kwargs = [
        'safe_channels_path',
        'feature_columns',
        'window',
        'time',
        'significance',
        'layers',
        'activationfunction',
        'numnodes',
        'loss',
        'optimizer',
        'metrics',
        'epochs',
        'batchsize',
    ]

    def __init__(self, *args, **kwargs):
        base_classifiers.SupervisedClassifier.__init__(self, *args, **kwargs)

    def train(self, quiver):
        """
        Trains a model using feature data, makes predictions using the training set's class distribution
        """
        if not self.is_trained:
            self.model = self._define_model(quiver.start, quiver.end, segs=quiver.segs)

        data, names = quiver.vectorize(
            self.model.channels,
            self.model.select,
            time=self.model.time,
            num_proc=self.kwargs.get('num_proc', features.DEFAULT_NUM_PROC),
        )

        self.model.whitener.train(data, names)
        self.model.keras.fit(
            self.model.whitener.whiten(data, names),
            quiver.labels,
            epochs=self.kwargs['epochs'],
            batch_size=self.kwargs['batchsize'],
            verbose=1,
        )
        return self.model

    def evaluate(self, quiver):
        """
        Applies a supervised keras model to feature data.
        """
        if not self.is_trained:
            raise base_classifiers.UntrainedError('%s does not have an internal model'%self.flavor)

        data, names = quiver.vectorize(
            self.model.channels,
            self.model.select,
            time=self.model.time,
            num_proc=self.kwargs.get('num_proc', features.DEFAULT_NUM_PROC),
        )

        ranks=self.model.keras.predict(self.model.whitener.whiten(data, names), verbose=1)
        quiver.apply_ranks(ranks, hashes=self.model.hash)
        return quiver

    def timeseries(self, quiver_factory, dt=base_classifiers.DEFAULT_DT, segs=None):
        """
        Generate a time series of predictions based on predicted model probabilities.
        """
        ### check if model has been trained
        if not self.is_trained:
            raise base_classifiers.UntrainedError('%s does not have an internal model'%self.flavor)

        if segs is None:
            segs = quiver_factory.classifier_data.segs

        ranks = []
        for times in utils.segs2times(segs, dt):
            quiver = quiver_factory(times)
            if len(quiver): ### append ranks if quiver isn't empty
                ranks.append( (self.evaluate(quiver).ranks, times[0], dt) )
        return ranks

    def feature_importance(self):
        """
        return a ranked list of important features within the trained model

        WARNING: NOT IMPLEMENTED YET!
        """
        raise NotImplementedError

    def _define_model(self, start, end, segs=None):
        """
        Defines and returns a keras classifier model, using keyword arguments as necessary.
        Used internally
        """
        ### set defaults for missing values if not set
        self.kwargs['default'] = self.kwargs.get('default', DEFAULT_COLUMN_VALUE)
        self.kwargs['default_delta_time'] = self.kwargs.get('default_delta_time', self.kwargs['window'])

        kwargs = {
            'keras':self._define_keras(),
            'whitener':features.Whitener(),
            'channels':io.path2channels(self.kwargs['safe_channels_path']),
            'select':features.SelectLoudest(**self.kwargs),
            'time':self.kwargs['time']
        }
        return KerasModel(
            start,
            end,
            self._define_keras(),
            features.Whitener(),
            io.path2channels(self.kwargs['safe_channels_path']),
            features.SelectLoudest(**self.kwargs),
            time=self.kwargs['time'],
            segs=segs,
        )

    def _define_keras(self, *args, **kwargs):
        return KerasPickleWrapper(keras_sequential())

#------------------------
#   Neural Networks
#------------------------

class NeuralNetwork(SupervisedKerasClassifier):
    """
    A neural network based on Keras
    """
    _flavor = "neural_network"

    def _define_keras(self, *args, **kwargs):
        # FIXME This only supports Dense layers as currently written  
        model = keras_sequential()

        ### NOTE: input dimension = num_channels * num_columns
        num_inputs = len(io.path2channels(self.kwargs['safe_channels_path'])) * len(self.kwargs['feature_columns'])

        N = len(self.kwargs['layers'])
        assert N==len(self.kwargs['numnodes']), "len(numnodes) != len(layers)"
        assert N==len(self.kwargs['activationfunction']), "len(activationfunction) != len(layers)"

        for i, (layer, num_nodes, activation) in enumerate(zip(self.kwargs['layers'], self.kwargs['numnodes'], self.kwargs['activationfunction'])):
            if i==0:
                if layer=='Dense':
                    model.add(keras_dense(units=num_nodes, activation=activation, input_dim=num_inputs))
                else:
                    raise ValueError('layer=%s not understood when it is the first layer'%layer)

            else:
                if layer=='Dense':
                    model.add(keras_dense(units=num_nodes, activation=activation))
                else:
                    raise ValueError('layer=%s not understood'%layer)

        model.compile(loss=self.kwargs['loss'], optimizer=self.kwargs['optimizer'], metrics=self.kwargs['metrics'])
        return KerasPickleWrapper(model)
