__description__ = "a module that houses routines to automate instantiation of objects"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import numpy as np

### non-standard libraries
from ligo import segments as ligo_segments

from idq import classifiers
from idq import features
from idq import io
from idq import utils

#-------------------------------------------------

DEFAULT_DT = classifiers.DEFAULT_DT

#-------------------------------------------------
# things from io

class ClassifierDataFactory(object):
    """
    A factory to create instances of ClassifierData-like objects
    """
    def __init__(self):
        self.known_subclasses = {cls.__name__: cls for cls in utils.all_subclasses(io.ClassifierData)}

    def __call__(self, start, end, segs=None, flavor=None, **kwargs):
        """
        standardizes how we map classifier parameters into an actual object

          * start is a float (gps seconds)
          * end is a float (gps seconds)
          * segs, if provided, should be a list of segments (a list of 2-element lists)
          * kwargs should be the result of **config.items(nickname)

        NOTE: 'flavor' is a reserved option that must be supplied

        returns

          * a classifier_data object corresponding to the parameters within **kwargs
        """
        try:
            return self.known_subclasses[flavor](start, end, segs=segs, **kwargs)

        except KeyError:
            raise KeyError, '%s is not a known ClassifierData!' %flavor ### format error reporting a bit better

#------------------------

class ReporterFactory(object):
    """
    A factory to create instances of Reporter-like objects
    """
    def __init__(self):
        self.known_subclasses = {cls.__name__: cls for cls in utils.all_subclasses(io.Reporter)}

    def __call__(self, rootdir, start, end, flavor=None, **kwargs):
        """
        standarize how we construct reporters
        """
        try:
            return self.known_subclasses[flavor](rootdir, start, end, **kwargs)

        except KeyError:
            raise KeyError, '%s is not a known Reporter!' %flavor ### format error reporting a bit better

#-------------------------------------------------
# things from classifiers

class ClassifierFactory(object):
    """
    A factory to create instances of SupervisedClassifier-like objects
    """
    def __init__(self):
        self.known_subclasses = {cls.__name__: cls for cls in utils.all_subclasses(classifiers.SupervisedClassifier)}

    def __call__(self, nickname, flavor=None, **kwargs):
        try:
            return self.known_subclasses[flavor](nickname, **kwargs)

        except KeyError:
            raise KeyError, '%s is not a known SupervisedClassifier!'%flavor ### format error reporting a bit more nicely


class IncrementalClassifierFactory(object):
    """
    A factory to create instances of IncrementalSupervisedClassifier-like objects
    """
    def __init__(self):
        self.known_subclasses = {cls.__name__: cls for cls in utils.all_subclasses(classifiers.IncrementalSupervisedClassifier)}

    def __call__(self, nickname, flavor=None, **kwargs):
        try:
            return self.known_subclasses[flavor](nickname, **kwargs)

        except KeyError:
            raise KeyError, '%s is not a known IncrementalSupervisedClassifier!'%flavor ### format error reporting

#-------------------------------------------------
# things from features

class QuiverFactory(object):
    """
    a factory object to generate quivers
    """

    def __init__(self, classifier_data):
        self.classifier_data = classifier_data

    def labeled(self, target_times, random_times):
        """
        construct a quiver based on classifier_data at these times
            target_times -> label=1.
            random_times -> label=0.
        """
        return self(
            np.concatenate((target_times, random_times)),
            labels=[1.]*len(target_times) + [0.]*len(random_times),
        )

    def unlabeled(self, dt=DEFAULT_DT, segs=None):
        """
        construct a quiver based on classifier_data at a regular time interval dt.
        """
        if segs is None:
            segs = self.classifier_data.segs
        return self(np.concatenate(tuple(utils.segs2times(segs, dt))))

    def __call__(self, times, labels=None):
        """
        a helper function to standardize how we assemble bits of quivers
        """
        if labels is None: ### set up labels if they aren't passed
            labels = [None]*len(times)
        else:
            assert len(labels)==len(times), 'labels and times must have the same length!'

        classifier_data = self.classifier_data ### pull out this reference once for convenience
        pairs = None ### tuples of (gps, label, classifier_data) later used to instantiate FeatureVectors
                     ### this may be overwritten within the following conditionals, otherwise we default to "standard behavior"

        ### we're not told to skip the fancy stuff and we have enough information to do at least some fancy stuff
        if classifier_data.kwargs.has_key("window") and (not classifier_data.kwargs.get("direct", False)):

            ### condition on whether we filter segments around times
            ### NOTE: this creates a new object, so it will not create a shared reference pointing to the classifier_data that was input
            if not classifier_data.kwargs.get('do_not_window', False):

                classifier_data = ClassifierDataFactory()( ### don't hold onto a reference because we only use it once in this conditional...
                    classifier_data.start,
                    classifier_data.end,
                    segs=utils.segments_intersection(
                        classifier_data.segs,
                        utils.times2segments(times, classifier_data.kwargs['window']), ### FIXME: may want some logic that says when we should skip this additional filtering (if densely spaced samples are very dense, this could slow down I/O)
                    ),
                    flavor=type(classifier_data).__name__,
                    columns=classifier_data.columns,
                    **classifier_data.kwargs
                )

            ### condition on whether we divide I/O into small chunks by defining umbrellas
            if classifier_data.kwargs.has_key("cache") or classifier_data.kwargs.has_key("stride"):

                flavor = type(classifier_data).__name__
                columns = classifier_data.columns
                window = classifier_data.kwargs['window']

                ### divide I/O based on stride if available, default to cache segment boundaries otherwise
                if classifier_data.kwargs.has_key("stride"):
                    file_segs = sorted(utils.split_segments_by_stride(classifier_data.start, classifier_data.end, classifier_data.segs, classifier_data.kwargs['stride']))
                else:
                    file_segs = sorted(utils.cache2unique_segments(classifier_data.kwargs['cache']))

                ### generate a classifier data for each file span
                classifier_datum = {}
                classifier_data_kwargs = dict(classifier_data.kwargs.items())
                factory = ClassifierDataFactory() ### instantiate this because we'll use it to make new objects
                for file_seg in file_segs:
                    segs = utils.segments_intersection(classifier_data.segs, ligo_segments.segmentlist([file_seg]))

                    ### filter cache if classifier data uses one
                    if classifier_data.kwargs.has_key("cache"):
                        classifier_data_kwargs['cache'] = [entry for entry in classifier_data.kwargs['cache'] if segs.intersects_segment(entry.segment)]

                    classifier_datum[file_seg] = factory(
                        float(file_seg[0]),
                        float(file_seg[1]),
                        segs=segs,
                        flavor=flavor,
                        columns=columns,
                        **classifier_data_kwargs
                    )

                ### create an umbrella for each feature vector
                pairs = [] ### create iterable to store these tuples
                for gps, label in zip(times, labels):

                    ### find relevant classifier datum based on gps window
                    seg = ligo_segments.segment(gps-window, gps+window)
                    target_segs = utils.find_unique_segs(seg, file_segs)
                    if len(target_segs) == 0: ### no data for this gps window
                        continue

                    if len(target_segs) == 1:
                        umbrella_segs = classifier_datum[target_segs[0]].segs

                    else: ### gps window spans over two classifier datum
                        umbrella_segs = classifier_datum[target_segs[0]].segs | classifier_datum[target_segs[1]].segs

                    umbrella = io.UmbrellaClassifierData(
                        target_segs[0][0],
                        target_segs[-1][1],
                        segs=umbrella_segs,
                        columns=columns,
                        **classifier_data.kwargs
                    )

                    ### add classifier datum to umbrella
                    for target_seg in target_segs:
                        umbrella.extend_child(classifier_datum[target_seg])

                    pairs.append((gps, label, umbrella))

        if pairs is None: ### this was not overwritten by any of the conditionals
            pairs = [(gps, label, classifier_data) for gps, label in zip(times, labels)]

        ### create resulting quiver, add in all the pairs defined above
        quiver = features.Quiver()
        for gps, label, obj in pairs:
            quiver.append(features.FeatureVector(gps, label, obj))

        quiver.start = classifier_data.start
        quiver.end = classifier_data.end
        quiver.segs = self.classifier_data.segs

        return quiver
