__description__ = "a python module housing basic iDQ functionality"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import glob
import json
import os
import pickle
import shutil
import time
import timeit

import h5py
import numpy as np

from collections import defaultdict
from collections import deque

import subprocess as sp ### only used as a work-around for reading FrameHistory from disk

### non-standard Python modules
from lal import CreateREAL8TimeSeries
from lal import CreateREAL8Sequence
from lal import DimensionlessUnit
from lal import LALDETECTORTYPE_ABSENT

from LDAStools import frameCPP as frcpp

from ligo import segments as ligo_segments
from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils as ligolw_utils
from glue.ligolw.utils import segments as ligolw_segments

from idq import names
from idq import utils
from idq import features
from idq import calibration
from idq import synthetic
from idq import configparser

from idq import exceptions

#-------------------------------------------------

DEFAULT_STREAMREPORTER_TIMEOUT = 0.2
DEFAULT_SHMDIR_AGE = 600

#-------------------------------------------------
### misc utility functions

def path2channels(path):
    """
    extracts a list of channels from a path
    expects each row to be a single channel name
    returns a list
    """
    with open(path, 'r') as file_obj:
        channels = [_.strip() for _ in file_obj.readlines()]

    return channels

#-------------------------------------------------

"""
need to define data object that we'll use throughout. Should be similar to trgDict from the old pipeline.
    Just define the class here (along with its attributes) and then use it in, e.g., ~lib/classifiers.py and ~lib/features.py
    data object should contain a reference to segments which it respresents (start,stop tuples, DQ flag name, etc). In this way, we can pass one object to both "discrete sample" classifiers, which will ignore the segments, and pass the same object to "poisson rate" classifiers, which need to know how much time was associated with the sample set.
"""

class ClassifierData(object):
    """
    an object which represents the "data" used within classifiers.
    specifically, this object should define getters for each feature that could be requested and store them as @property values.
    In this way, we can encapsulate all the interactions with (S)MDS within this class and maintain a single API with classifiers even if (S)MDS changes (e.g. filesystem query vs remote data base query)

    Note, this is meant as a parent class and specific look-up can be defined through children
    """
    _default_columns = [] ### which columns will be pulled from triggers by default
    _allowed_columns = None ### which columns are allowed; None allows anything
    _required_kwargs = [] ### which kwargs are required for this type of data object

    def __init__(self, start, end, segs=None, columns=None, **kwargs):
        """
        each instance points to (what should be) immutable
            - segments
            - columns
        and when data is queried through .triggers() it will automatically be within these segments and only those columns
        In particular, data is stored internally as a dictionary of the following form
            (channel, numpy.ndarray([np.ndarray([col1_val, col2_val, ...]), ...]))
        the values within the array are ordered according to columns

        *WRITE ME*
            describe attribute inheritence and rational for these
                _default_columns (I want to make each child class declare these for themselves so the developer knows what he's doing)
                _allowed_columns
                _required_kwargs
                _data
                _start
                _end
                _segs
                kwargs
                _columns
            describe method inheritance, what each method is supposed to do
                triggers
                _retrieve_triggers (must be overwritten by child)
                pop
                __add__
        """
        self._data = {}

        assert start <= end, "start must be <= end"
        self._start = start
        self._end = end

        if segs is not None:
            assert utils.livetime(utils.segments_intersection([(self._start, self._end)], segs)) == utils.livetime(segs), '''\
segments must be a subset of [[start, end]] times'''

        else:
            segs = ligo_segments.segmentlist([ligo_segments.segment(start, end)])
        self._segs = segs

        for kwarg in self._required_kwargs:
            assert kwargs.has_key(kwarg), 'kwarg=%s required'%kwarg
        self.kwargs = kwargs

        if not columns:
            columns = self._default_columns

        if self._allowed_columns is not None:
            for col in columns:
                assert col in self._allowed_columns, 'column=%s not allowed'%col
        self._columns = columns ### FIXME: make a copy instead of sharing the reference?

    def is_cached(self, channel):
        """
        Returns whether or not data is cached for the channel requested.
        """
        return self._data.has_key(channel)

    def triggers(self, channels=None, columns=None, segs=None, time=None, **kwargs):
        """
        Submits a query to get features and returns the result of such a query,
        as a dictionary of the following form:

            {channel: numpy.ndarray([np.ndarray([col1_val, col2_val, ...]), ...])}

        The values within the array are ordered according to columns.
        """

        ### format and validate query
        if channels:
            if isinstance(channels, str):
                channels = [channels]

        if segs:
            assert segs in self.segs, 'segments=%s passed in query are not a subset of classifier data segments'%segs
            assert time, 'if passing in segments, time column needs to be specified'

        if columns and (self._allowed_columns is not None):
            if isinstance(columns, str):
                columns = [columns]
            for col in columns:
                assert col in self._allowed_columns, 'column=%s not allowed'%col

        ### determine if there's any cached data
        if self._data and (channels is not None):
            channels2query = [chan for chan in channels if not self.is_cached(chan)]
            if channels2query:
                self._data.update(self._retrieve_triggers(channels=channels2query, **kwargs))

        else: ### no cached data, retrieve all features from query
            self._data.update(self._retrieve_triggers(channels=channels, **kwargs))

        ### grab queried data from cache
        if channels:
            queried_data = {channel: self._data[channel] for channel in channels}
        else:
            queried_data = dict(self._data)

        if segs:
            for channel in queried_data.keys():
                times2query = utils.times_in_segments(queried_data[channel][time], segs)
                queried_data[channel] = queried_data[channel][times2query]

        if columns:
            for channel in queried_data.keys():
                queried_data[channel] = queried_data[channel][columns]

        return queried_data

    @property
    def children(self):
        """
        Makes an iteration over both ClassifierDatas and UmbrellaClassifierDatas a little easier.

        WARNING: Users shouldn't mess with this unless they know what they're doing.
        """
        return [self]

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def segs(self):
        return self._segs

    @property
    def columns(self):
        return self._columns

    @property
    def channels(self):
        return self._data.keys()

    def restrict_segments(self, new_segs, time=features.DEFAULT_TIME_NAME):
        """
        update the internal segments to new_segs (checks that new_segs is a strict subset of self.segs)
        iterates through cached triggers and throws away everything outside of new_segs
        NOTE: this requires knowledge of the "time" key within triggers
        """
        assert utils.livetime(utils.segments_intersection(self.segs, new_segs)) == utils.livetime(new_segs), 'new_segs must be a strict subset of existing segs'
        self._segs = ligo_segments.segmentlist([ligo_segments.segment(s, e) for s, e in new_segs]) ### make a copy so there is no cross-refernce confusion
        for channel in self._data.keys():
            trgs = self._data[channel]
            truth = utils.times_in_segments(trgs[time], self.segs)
            self._data[channel] = trgs[truth]

    def _retrieve_triggers(self, **kwargs):
        raise NotImplementedError, 'Child classes should define this for themselves!'

    def pop(self, channel, *args):
        """
        Remove (and return) the value associated with channel in the internal structure.

        NOTE: once this is called, object will perform a new query if channel is requested again via triggers.
        """
        return self._data.pop(channel, *args)

    def __add__(self, other):
        """
        return a NEW OBJECT that represents the addition of 2 data objects
        we take the union of segments and columns
        we also keep the union of channels from each object.
        if data was already present in either object, we go ahead and query for any missing data at the time of the addition
        """
        raise NotImplementedError, "this could be useful, but could also be difficult to implement. For instance, we need to make sure the children instantiate the correct type of object "

    def target_times(self, time, target_channel, target_bounds, segs=None):
        """
        A convenience function to extract target_times from umbrellas in streaming-based processes.
        """
        ### find target times
        target_times = utils.filter_triggers(self.triggers(channels=target_channel)[target_channel], bounds=target_bounds)[time]

        ### filter times by segments
        if segs is not None:
            target_times = target_times[utils.times_in_segments(target_times, segs)]

        return target_times

    def random_times(self, time, target_channel, dirty_bounds, dirty_window, random_rate, segs=None):
        """
        A convenience function to extract target_times from umbrellas in streaming-based processes.
        """
        ### draw random times
        dirty_times = utils.filter_triggers(self.triggers(channels=target_channel)[target_channel], bounds=dirty_bounds)[time]
        dirty_seg = utils.times2segments(dirty_times, dirty_window) ### generate segments

        ### only draw from segments that are outside of dirtysegs but within classifier_data.segs
        data_segs = ligo_segments.segmentlist([ligo_segments.segment(self.start, self.end)])
        random_times = utils.draw_random_times(utils.remove_segments(data_segs, dirty_seg), rate=random_rate)

        ### filter times by segments
        if segs is not None:
            random_times = random_times[utils.times_in_segments(random_times, segs)]

        return random_times

class UmbrellaClassifierData(ClassifierData):
    """
    a wrapper around a list of ClassifierData objects, useful when trying to not repeat I/O but splitting it into many small chunks
    """
    def __init__(self, *args, **kwargs):
        ClassifierData.__init__(self, *args, **kwargs)
        self._children = deque(maxlen=kwargs.get('max_children', None))
        self._channels = set()

    @property
    def all_channels(self):
        """
        return all channels present in any child
        """
        chans = set()
        for child in self.children:
            chans |= set(child.channels)
        return list(chans)

    @property
    def columns(self):
        """
        return the columns that are common to all children
        """
        n = len(self.children)
        if n:
            cols = set(self.children[0].columns)
            i = 1
            while i < n:
                cols &= set(self.children[i].columns)
                i += 1
            return list(cols)
        else:
            return []

    def all_columns(self):
        """
        return all columns present in any child
        """
        cols = set()
        for child in self.children:
            cols |= set(child.columns)
        return list(cols)

    @property
    def children(self):
        return self._children

    @property
    def channels(self):
        return list(self._channels)

    def _check_child(self, child):
        """
        make sure the child is not requesting any data not covered by the Umbrella
        """
        lvtm1 = utils.livetime(utils.segments_intersection(self.segs, child.segs))
        lvtm2 = utils.livetime(child.segs)
        assert lvtm1==lvtm2, 'livetime from "self.segs AND child.segs" (%.6f) is not the same as "child.segs" (%.6f)\n%s\n%s'%(lvtm1, lvtm2, utils.segments_intersection(self.segs, child.segs), child.segs)
        ### NOTE:
        ###     Fe could check a lot more there.
        ###     For example, we could check that no children overlap, but we may want children to overlap at some point
        ###         (e.g.: read in both KW triggers and GSTLAL triggers)
        ###     we may also want a check that the full set of children included covers the requested range, but that seems like we should pass children at instantiation time rather than appending htem later. Could be a useful safety net, but not necessarily what we want

    def _update_channels(self, child):
        """
        update Umbrella with common set of channels covered
        """
        if not self._channels:
            self._channels = set(child.channels)
        else:
            self._channels &= set(child.channels)

    def extend_child(self, child):
        """
        modifies self.start, self.end in addition to adding the child
        allways appends child
        """
        self._start = min(self.start, child.start)
        self._end = max(self.end, child.end)
        self._segs = self._segs | child.segs
        self._update_channels(child)
        self._children.append(child)

    def extend_children(self, children):
        """
        modifies self.start, self.end in addition to adding the child
        always appends child
        """
        for child in children:
            self.extend_child(child)

    def strip_child(self):
        """
        removes the left-most child and updates start, end, segs accordingly
        """
        self._children.popleft()
        if len(self.children):
            self._start = min([child.start for child in self._children])
            self._end = max([child.end for child in self._children])
            self._segs = utils.segments_union(*[child.segs for child in self._children])
            self._channels = set()
            for child in self._children:
                self._channels &= set(child.channels)
        else:
            self._start = self._end ### we've removed the last child from the left, so we collapse the segment accordingly
            self._segs = ligo_segments.segmentlist([])
            self._channels = set()

    def restrict_segments(self, new_segs, time=features.DEFAULT_TIME_NAME):
        for child in self.children:
            child.restrict_segments(utils.segments_intersection(new_segs, child.segs), time=time)

    def triggers(self, channels=None, columns=None, segs=None, time=None, **kwargs):
        self._data = {}
        for child in self.children: ### do I/O for each child
            for key, val in child.triggers(channels=channels, columns=columns, segs=segs, time=time, **kwargs).items():
                if self._data.has_key(key):
                    self._data[key] = np.concatenate((self._data[key], val))

                else:
                    self._data[key] = val

            self._update_channels(child)

        return self._data

#------------------------
### synthetic data

class MockClassifierData(ClassifierData):
    """
    a synthetic data generation object
    """
    _default_columns = synthetic.MOCKCLASSIFIER_COLUMNS
    _allowed_columns = synthetic.MOCKCLASSIFIER_COLUMNS
    _required_kwargs = ['config']

    _default_prob_nodata = 0.
    _default_prob_incontiguous = 0.

    def __init__(self, *args, **kwargs):
        ClassifierData.__init__(self, *args, **kwargs)
        self._parse_config(self.kwargs['config'])
        self._dtype = [(key, float) for key in self.columns]

        if not self.kwargs.has_key('prob_nodata'):
            self.kwargs['prob_nodata'] = self._default_prob_nodata
        if not self.kwargs.has_key('prob_incontiguous'):
            self.kwargs['prob_incontiguous'] = self._default_prob_incontiguous

    def _parse_config(self, path):
        """
        read in parameters from config by delegating to idq.synthetic.parse_mockclassifierdata_config
        """
        self._streams, self._channel2streams = configparser.path2streams(path, self.segs)

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """
        generates synthetic data on-the-fly based on parameters read from config file
        generates and caches data for all channels declared in config instead of trying to only generate the channels requested
        """
        if self._data:
            synthetic_data = self._data ### handle caching explicitly here because otherwise calls to triggers() without specifying channels
                                        ### will re-generate the random streams. This is special behavior just for the Synthetic data objects
                                        ### so we take care of it here rather than in the parent class
                                        ### relies on the fact that we generate all data for all channels the first time this is called

        else: ### generate all streams of synthetic triggers

            ### figure out if we want to raise an error
            if np.random.rand() < self.kwargs['prob_nodata']: ### raise a NoDataError
                raise exceptions.NoDataError(self.start, self.end-self.start)

            if np.random.rand() < self.kwargs['prob_incontiguous']: ### raise BadSpanError
                gap_start = self.start+np.random.rand()*(self.end-self.start)
                gap_end = gap_start + np.random.rand()(self.end-gap_start)
                raise exceptions.IncontiguousDataError(gap_start, gap_end, self.end)

            ### no errors raised, so we simulate streams
            all_streams = dict((obj.name, obj) for obj in self._streams)

            ### map those streams into the appropriate witnesses
            synthetic_data = {}
            for channel, streams in self._channel2streams.items():
                arrays = [np.array([], dtype=self._dtype)] ### incrementally add streams to this for concatenation at the end
                for stream, jitters in streams: ### add jitter to each observed stream
                    arrays.append(stream.jittered(**jitters)[self.columns]) ### add jitter and downselect to the columns requested
                synthetic_data[channel] = np.concatenate(tuple(arrays))

            if synthetic.MOCKCLASSIFIER_TIME in self.columns: ### time-order the data if we've requested the time column
                for channel, val in synthetic_data.items():
                    synthetic_data[channel] = val[val[synthetic.MOCKCLASSIFIER_TIME].argsort()]

        if channels is not None:
            for chan in channels: ### fill in missing channels
                if not synthetic_data.has_key(chan):
                    synthetic_data[chan] = np.array([], dtype=self._dtype)
        return synthetic_data

class DynamicMockClassifierData(UmbrellaClassifierData):
    """
    a synthetic data generation object that allows for multiple correlation configs
    """
    _default_columns = synthetic.MOCKCLASSIFIER_COLUMNS
    _allowed_columns = synthetic.MOCKCLASSIFIER_COLUMNS
    _required_kwargs = ['configs']

    def __init__(self, *args, **kwargs):
        UmbrellaClassifierData.__init__(self, *args, **kwargs)

        ### make children as needed
        self.kwargs['configs'] = configparser.config2bounds(self.kwargs['configs'])
        ans = [(ligo_segments.segmentlist([ligo_segments.segment(s, e)]), path) for path, (s, e) in self.kwargs['configs'].items()]
        ans.sort(key=lambda _: _[0][0][0])
        for i, (segs1, path1) in enumerate(ans):
            for segs2, path2 in ans[i+1:]:
                if utils.livetime(utils.segments_intersection(segs1, segs2)):
                    raise ValueError('overlapping segments for %s [%.3f, %.3f) and %s [%.3f, %.3f)'%(path1, segs1[0][0], segs1[0][1], path2, segs2[0][0], segs2[0][1]))

        for segments, path in ans: ### create children as needed
            intersection = utils.segments_intersection(self.segs, segments)
            if utils.livetime(intersection): ### this child is relevant, so we include it
                start = intersection[0][0]
                end = intersection[-1][1]
                self.extend_child(MockClassifierData(start, end, segs=intersection, config=path, **self.kwargs))
            
#------------------------
### KleineWelle methods and objects
### objects inherit from ClassifierData

def _kws2triggers(path, segments=None, columns=utils.KW_COLUMNS):
    if segments is None:
        segments = [[-np.infty, +np.infty]] ### grab everything

    data = []
    num_segs = len(segments)
    if num_segs == 0:
        return data

    segind = 0
    inds = [utils.KW_COL2IND[col] for col in columns] ### which indecies we extract from each line

    with open(path, 'r') as file_obj:
        start, end = segments[segind] ### which segment we're working with

        for line in file_obj:
            if line[0] in utils.ASCII_COMMENTS: ### commented
                continue

            fields = _kws_line2fields(line) ### parse the line
            t = fields[utils.KW_TIMEIND]

            # handle segment logic here in an attempt to speed things up (we assume sorting here)
            if t > end: ### trigger is too late, bump the segment
                segind += 1
                while segind < num_segs:
                    if t > segments[segind][1]:
                        segind += 1
                    else:
                        break
                else:
                    break ### did not find another segment, so we're done with this file
                start, end = segments[segind] # update definition of start, end

            if t < start: ### trigger is too early, but we still have segments, so go to next trigger
                continue

            ### now, start < t < end is guaranteed, so we add this segment
            data.append( tuple(fields[inds]) )

    return data

def _kws_line2fields(line):
    return np.array([float(val) for val in line.strip().split()]) ### NOTE: this is faster than np.array(...).astype('float')

def _kwm2triggers(path, segments=None, columns=utils.KW_COLUMNS, channels=None):
    if segments is None:
        segments = [[-np.infty, +np.infty]]

    data = defaultdict(list)
    num_segs = len(segments)
    if num_segs == 0:
        return data

    segind = 0
    inds = [utils.KW_COL2IND[col] for col in columns] ### which indecies we extract from each line

    with open(path, 'r') as file_obj:
        start, end = segments[segind]

        for line in file_obj:
            if line[0] in utils.ASCII_COMMENTS:
                continue

            fields, chan = _kwm_line2fields(line)
            if channels and (chan not in channels): ### we don't care about this channel
                continue

            # we do care about this channel, so now we check segments
            t = fields[utils.KW_TIMEIND]

            # handle segment logic here in an attempt to speed things up (we assume sorting here)
            if t > end: ### trigger is too late, bump the segment
                segind += 1
                while segind < num_segs:
                    if t > segments[segind][1]:
                        segind += 1
                    else:
                        break
                else:
                    break ### did not find another segment, so we're done with this file
                start, end = segments[segind] # update definition of start, end

            if t < start: ### trigger is too early, but we still have segments, so go to next trigger
                continue

            ### now, start < t < end is guaranteed, so we add this segment
            data[chan].append( tuple(fields[inds]) )

    return data

def _kwm_line2fields(line):
    fields = line.strip().split()
    return np.array([float(val) for val in fields[:-1]]), fields[-1]

#-----------

class PredictiveKWSClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in single-channel files organized in a directory structure with a constant (known) stride
    Note: if we do not request specific channels, all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir', 'stride']

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for KWS files
        predicts directory names through expected directory structure and then looks for channel.trg files within each directory as needed
        we assume the following directory structure
            ${rootdir}/${gpsstart}_${gpsstop}/${channel}.trg
        """
        if verbose:
            print( '''\
searching for KWSingle triggers within
    rootdir : %(rootdir)s
with
    stride : %(stride)d'''%self.kwargs)
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print('within segments')
            for seg in self.segs:
                print('    %.3f - %.3f'%tuple(seg) )

        # extract things that will be used repeatedly
        stride = self.kwargs['stride']

        ### iterate through segments, attaching directories to each segment as needed
        template = names.start_end2dirtemplate(rootdir=self.kwargs['rootdir'])
        directories = defaultdict(list)
        for s, e in self.segs:
            S = (int(s)/stride)*stride ### integer division to get the directory starting point!
            while S < e:
                E = S+stride
                directory = template%(S, E)
                if os.path.exists(directory):
                    directories[directory].append( (s, min(e,E)) )
                S = E ### advance by one stride

        ### now actually read in triggers
        data = defaultdict(list) ### data structure for query results
        if channels is not None:
            chan_paths = [(chan, names.kws_channel2filename(chan)) for chan in channels] ### define this for convenience later

        # iterate through directories that exist
        for directory in sorted(directories.keys()): ### make sure these are sorted
            if verbose:
                print( '    '+directory )

            relevant_segments = utils.fix_segments(directories[directory]) ### extract and fix segments for this directory
            if verbose:
                for seg in relevant_segments:
                    print( '        %.3f - %.3f'%tuple(seg) )

            ### get relevant paths for this directory
            ### the resulting list is guaranteed to only contain paths that exist
            if channels is not None:
                paths = []
                for chan, chan_path in chan_paths:
                    path = os.path.join(directory, chan_path)
                    if os.path.exists(path):
                        paths.append((chan, path))

            else:
                paths = [(os.path.basename(path)[:-4], path) for path in glob.iglob(os.path.join(directory, '*.trg'))]

            # iterate through paths, reading in triggers for each path
            for chan, path in paths:
                if verbose:
                    print( '        '+path )
                data[chan] += _kws2triggers(path, segments=relevant_segments, columns=self.columns)

        ### clean crap up
        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels is not None:
            for chan in channels: ### fill in missing channels
                if not data.has_key(chan):
                    data[chan] = np.array([], dtype=dtype)

        return data

#-----------

class KWSClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in single-channel files
    Note: if we do not request specific channels, all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir']

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for KWS files
        globs through expected directory structure only once and then looks for channel.trg files within each directory as needed
        we assume the following directory structure
            ${rootdir}/${gpsstart}_${gpsstop}/${channel}.trg
        """
        if verbose:
            print( '''\
searching for KWSingle triggers within
    rootdir : %(rootdir)s'''%self.kwargs )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    %.3f - %.3f'%tuple(seg) )

#        data = dict((chan,[]) for chan in channels) ### data structure for query results
        data = defaultdict(list) ### data structure for query results

        if channels is not None:
            chan_paths = [(chan, names.kws_channel2filename(chan)) for chan in channels] ### define this for convenience later

        ### find which directories contain the stretches we want
        for directory in glob.iglob(os.path.join(self.kwargs['rootdir'], '*_*/')):
            if verbose:
                print( '    '+directory )

#            dir_seg = names.dir2start_end(directory)
            dir_seg = ligo_segments.segment(names.dir2start_end(directory))

            ### determine which segments have any overlap with this directory
            ### will save time when we filter triggers specifically from this directory
#            relevant_segments = []
            relevant_segments = ligo_segments.segmentlist()
            for seg in self.segs: ### check for overlap with any segments
                if utils.segments_overlap(dir_seg, seg): ### non-zero overlap
                    relevant_segments.append( seg )

            if not relevant_segments: ### no overlap
                continue

            if verbose:
                for seg in relevant_segments:
                    print( '        %.3f - %.3f'%tuple(seg) )

            if channels is not None: ### only load a specific list
                paths = []
                for chan, chan_path in chan_paths:
                    path = os.path.join(directory, chan_path) ### predict the exact path and
                                                              ### require an exact match
                    if os.path.exists(path):
                        paths.append((chan, path))

            else: ### load everything you can find
                paths = [(os.path.basename(path)[:-4], path) for path in glob.iglob(os.path.join(directory, '*.trg'))]

            for chan, path in paths:
                if verbose:
                    print( '        '+path )
                data[chan] += _kws2triggers(path, segments=relevant_segments, columns=self.columns)

        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels is not None:
            for chan in channels: ### fill in missing channels
                if not data.has_key(chan):
                    data[chan] = np.array([], dtype=dtype)

        return data

#-----------

class PredictiveKWMClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in multi-channel files with a known stride

    Note, if we do not request any specific channel(s), all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir', 'basename', 'stride']

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """
        read all relevant triggers from a single KWM file in at once to save on I/O
        we predict filenames and assume the following directory structure
            ${rootdir}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.trg
        """
        if verbose:
            print( '''\
searching for KWMultiple triggers within
    rootdir : %(rootdir)s
with
    basename : %(basename)s
    stride   : %(stride)d'''%self.kwargs )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    %.3f - %.3f'%tuple(seg) )

        ### pull out things referenced more than once
        stride = self.kwargs['stride']
        basename = self.kwargs['basename']

        ### iterate through directories, discovering which paths are relevant
        paths = defaultdict(list)

        template = names.start2dirMOD1e5_template(basename=self.kwargs['basename'], rootdir=self.kwargs['rootdir'])

        for s, e in self.segs:
            ints = int(s)
            Smod5 = (ints/100000) # integer division; hard-coded because this is what KWM always produces
            S = Smod5*100000

            while S < e: ### iterate over directories
                E = S+100000
                end = min(E, e) # the relevant end time for this directory
                SS = (max(ints, S)/stride)*stride ### the relevant starting point for this directory

                path_template = os.path.join(template%Smod5, names.kwm_stridebasename2filename(stride=stride, basename=basename))
                while SS < end: ### iterate over files in this directory
                    EE = SS+stride
                    path = path_template%SS

                    if os.path.exists(path):
                        paths[path].append( (SS, min(EE, e)) )
                    SS = EE # bump file start time by one stride

                S = E # bump directory start time by one stride (100000%100000 = 1)

        ### iterate over paths and extract data
        data = defaultdict(list) ### data structure for query results
        for path in sorted(paths.keys()): ### load these in in order
            if verbose:
                print(path)

            relevant_segments = utils.fix_segments(paths[path]) ### extract the useful segments
            if verbose:
                for seg in relevant_segments:
                    print('    %.3f - %.3f'%tuple(seg))

            for channel, triggers in _kwm2triggers(path, segments=relevant_segments, columns=self.columns, channels=channels).items():
                data[channel] += triggers

        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels:
            for chan in channels: ### fill in missing channels
                if not data.has_key(chan):
                    data[chan] = np.array([], dtype=dtype)

        return data

#-----------

class KWMClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in multi-channel files

    Note, if we do not request any specific channel(s), all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir', 'basename']

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """
        read all relevant triggers from a single KWM file in at once to save on I/O
        we assume the following directory structure
            ${rootdir}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.trg
        """
        if verbose:
            print( '''\
searching for KWMultiple triggers within
    rootdir : %(rootdir)s
with
    basename : %(basename)s'''%self.kwargs )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    %.3f - %.3f'%tuple(seg) )

#        data = dict((chan,[]) for chan in channels) ### data structure for query results
        data = defaultdict(list) ### data structure for query results

        basename_template = self.kwargs['basename']+'-*.trg' ### define this once, will be used repeatedly below

        ### find which directories contain the stretches we want
        for directory in glob.iglob(os.path.join(self.kwargs['rootdir'], self.kwargs['basename']+'-*/')):
            if verbose:
                print( '    '+directory )

#            dir_seg = names.dirMOD1e52start_end(directory)
            dir_seg = ligo_segments.segment(names.dirMOD1e52start_end(directory))

            ### determine which segments are useful for this directory
            ### will save time later on
#            relevant_segments = []
            relevant_segments = ligo_segments.segmentlist()
            for seg in self.segs: ### check for overlap with any segments
                if utils.segments_overlap(dir_seg, seg): ### non-zero overlap
                    relevant_segments.append( seg )

            if not relevant_segments: ### no overlap
                continue

            if verbose:
                for seg in relevant_segments:
                    print( '        %.3f - %.3f'%tuple(seg) )

            for trg_path in glob.iglob(os.path.join(directory, basename_template)): ### iterate over KWM files
#                trg_seg = names.path2start_end(trg_path, suffix='trg')
                trg_seg = ligo_segments.segment(names.path2start_end(trg_path, suffix='trg'))

#                germane_segments = [] ### same idea as relevant segments, but just for this trg file
                germane_segments = ligo_segments.segmentlist() ### same idea as relevant segments, but just for this trg file
                for seg in relevant_segments:
                    if utils.segments_overlap(trg_seg, seg):
                        germane_segments.append( seg )

                if not germane_segments: ### no overlap
                    continue

                if verbose:
                    print( '        '+trg_path )

                for channel, triggers in _kwm2triggers(trg_path, segments=germane_segments, columns=self.columns, channels=channels).items():
                    data[channel] += triggers

        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels:
            for chan in channels: ### fill in missing channels
                if not data.has_key(chan):
                    data[chan] = np.array([], dtype=dtype)

        return data

#-----------

class GstlalHDF5ClassifierData(ClassifierData):
    """
    An extension of ClassifierData specifically for extracting GstLAL features from hdf5 files.

    We assume the following directory structure:
        ${rootdir}/${basename}/${basename}-${gpsMOD1e5}/${basename}-${job_id}/${basename}-${start}-${dur}.h5

    """
    _default_columns = ['trigger_time', 'snr']
    _allowed_columns = utils.GSTLAL_COLUMNS
    _required_kwargs = ['rootdir', 'basename']

    def __init__(self, start, end, segs=None, columns=None, **kwargs):
        ClassifierData.__init__(self, start, end, segs=segs, columns=columns, **kwargs)

        ### load cache, or create one if one is not provided
        if 'cache' in self.kwargs and isinstance(self.kwargs['cache'], list):
            pass
        elif 'cache' in self.kwargs:
            self.kwargs['cache'] = utils.file2cache(self.kwargs['rootdir'], self.kwargs['cache'])
        else:
            path = names.basename2gstlal_filedir(self.kwargs['basename'])
            filename = names.basename2gstlal_filename(self.kwargs['basename'])
            self.kwargs['cache'] = utils.path2cache(self.kwargs['rootdir'], os.path.join(path, filename))

        ### filter cache with segments
        self.kwargs['cache'] = [entry for entry in self.kwargs['cache'] if self.segs.intersects_segment(entry.segment)]

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for gstlal hdf5 files
        we assume the following directory structure
            ${rootdir}/${basename}/${basename}-${gpsMOD1e5}/${basename}-${job_id}/${basename}-${start}-${dur}.h5
        """
        if verbose:
            print( '\n'.join([
                '',
                'searching for gstlal triggers within',
                '    rootdir : {rootdir}'.format(**self.kwargs),
                'with',
                '    basename : {basename}'.format(**self.kwargs),
                ])
            )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    {:3f} - {:3f}'.format(*seg) )

        ### data structure and dtype for query results
        data = defaultdict(list)
        dtype = [(col, '<f8') for col in self.columns]

        for entry in self.kwargs['cache']: ### iterate over hdf5 files

            ### determine which segments are relevant for this hdf5 file
            trg_seg = ligo_segments.segment(names.path2start_end(entry.path, suffix='h5'))
            germane_segments = utils.segments_intersection(self.segs, ligo_segments.segmentlist([trg_seg]))

            if verbose:
                print( '        '+entry.path )

            with h5py.File(entry.path, 'r') as file_obj:

                ### determine stride + sample rate of datasets
                for attr in ['cadence', 'sample_rate']:
                    assert attr in file_obj.attrs, '{} metadata not present in hdf5 file'.format(attr)
                cadence = file_obj.attrs.get('cadence')
                dataset_size = cadence * file_obj.attrs.get('sample_rate')


                ### query only channels in file
                file_channels = set(file_obj.keys())
                if channels is not None:
                    channels2query = file_channels & set(channels)
                else:
                    channels2query = file_channels - set(self.channels)

                ### set other properties defined inside a given hdf5 file
                possible_datasets = utils.segs2datasets(germane_segments, stride=cadence)

                for channel in channels2query:
                    datasets = set(file_obj[channel].keys())

                    ### filter datasets with segments
                    relevant_datasets = [pair for pair in possible_datasets if pair[0] in datasets]

                    if relevant_datasets:
                        key = os.path.join(channel, relevant_datasets[0][0])
                        row_dtype = file_obj[key].dtype

                        ### preallocate space and initialize all values to nan
                        rows = np.empty((dataset_size * len(relevant_datasets),), dtype=dtype)
                        rows[:] = np.nan

                        for dset_idx, pair in enumerate(relevant_datasets):
                            dataset, seglist = pair
                            key = os.path.join(channel, dataset)

                            if key in file_obj:
                                idx_start = dset_idx * dataset_size
                                idx_end = idx_start + dataset_size

                                # preallocate space for span of features wanted for a given dataset
                                triggers = np.empty((dataset_size,), dtype=dtype)
                                triggers[:] = np.nan

                                # create mask based on needed features
                                mask = np.ones(dataset_size, dtype=bool)
                                for seg in seglist:
                                    mask *= (float(seg[0])<=file_obj[key]['trigger_time'])*(file_obj[key]['trigger_time']<=float(seg[1]))

                                # filter features based on mask
                                triggers[mask] = file_obj[key][mask][self.columns]
                                rows[idx_start:idx_end] = triggers

                        nonnan_indices = np.where(np.isfinite(rows['trigger_time']))
                        data[channel].append(rows[nonnan_indices])

        ### stack datasets vertically
        for channel, datasets in data.items():
            data[channel] = np.concatenate(datasets, axis=0)

        ### fill in missing channels
        if channels:
            for channel in channels:
                if not data.has_key(channel):
                    data[channel] = np.array([], dtype=dtype)

        return data

#-----------

class KafkaClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for ingesting features from Kafka topics.

    Intended to keep a running current timestamp, and has the ability to poll for new data
    and fill its own ClassifierData objects for use when triggers are to be retrieved.
    """
    _default_columns = []
    _allowed_columns = None ### allow anything
    _required_kwargs = ['group', 'port', 'topic', 'poll_timeout', 'retry_cadence', 'sample_rate']

    def __init__(self, *args, **kwargs):
        ClassifierData.__init__(self, *args, **kwargs)

        ### link to Kafka consumer if one has not already been assigned
        if not self.kwargs.has_key('CONSUMER'):

            ### check if Kafka is installed
            utils.check_kafka()

            ### create a Kafka consumer
            self.kwargs['CONSUMER'] = utils.KafkaConsumer(
                self.kwargs['group'],
                self.kwargs['port'],
                topics=[self.kwargs['topic']],
                **self.kwargs.get('kafka_kwargs', {})
            )

            ### assign to specific partition to allow seeks/consumes
            self.kwargs['CONSUMER'].assign([
                utils.kafka.TopicPartition(self.kwargs['topic'], partition=self.kwargs['CONSUMER'].partition),
            ])

            ### FIXME: why do I need to get a message first before I can seek, even when assigning to a particular partition?
            msg = None
            while not msg:
                msg = self.kwargs['CONSUMER'].poll(timeout=self.kwargs['poll_timeout'])

        ### dtype for query results
        self._dtype = [(col, '<f8') for col in self.columns]

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        if not self._data: ### NOTE: this forces data to only be populated once

            if not utils.livetime(self.segs): ### we'll never find any data in the segments because they have zero livetime
                self._data = {}
                if channels is not None:
                    for channel in channels:
                        self._data[channel] = np.array([], dtype=self._dtype)
                return self._data

            ### there is the possibility of finding some data, so we go looking for that
            ### NOTE: assume last processed timestamp is pointed just before this span
            stride = 1. / self.kwargs['sample_rate']
            timestamp = self.start - stride
            segs = ligo_segments.segmentlist([ligo_segments.segment(self.start, self.end)]) ### the span of this object

            ### attempt to seek to correct offset
            try:
                self.kwargs['CONSUMER'].seek_timestamp(self.kwargs['topic'], self.start)
            except utils.kafka.KafkaException:
                pass

            ### check if timestamps fall within span of classifier data
            stop = self.end - stride
            while timestamp < stop:

                ### poll for data at requested offset
                start_time = timeit.default_timer()
                data_buffer = self.kwargs['CONSUMER'].poll(timeout=self.kwargs['poll_timeout'])

                if data_buffer and not data_buffer.error(): ### poll until we find something

                    ### process that something (incoming buffer)
                    data_buffer = json.loads(data_buffer.value())
                    new_timestamp = data_buffer['timestamp']
                    if verbose:
                        print('    timestamp of buffer: %.6f'%new_timestamp)
                        print('    latency: %.6f'%utils.gps2latency(new_timestamp))


                    if new_timestamp == timestamp+stride: ### data is contiguous, which is GOOD
                        if utils.time_in_segments(new_timestamp, self.segs): ### we care about this data
                            ### fill in relevant data from buffer
                            data = {}
                            for channel, rows in data_buffer['features'].items():
                                data[channel] = np.array([tuple(row[col] for col in self.columns) for row in rows if row], dtype=self._dtype)

                            ### add to existing triggers
                            for channel, triggers in data.items():
                                if self.is_cached(channel):
                                    self._data[channel] = np.concatenate((self._data[channel], triggers))
                                else:
                                    self._data[channel] = triggers

                        else:
                            pass ### we don't care about this data, so we just mosey along

                    else: ### data is NOT contiguous, which could be BAD
                        gap = ligo_segments.segment(timestamp+stride, new_timestamp)
                        ### check whether the gap touched things we care about, which is BAD
                        if utils.livetime(utils.segments_intersection(self.segs, ligo_segments.segmentlist([gap]))):
                            newer_timestamp = max(utils.floor_div(timestamp, self.end - self.start), self.start) ### should we retry or move to the next stride?
                            self.kwargs['CONSUMER'].seek_timestamp(self.kwargs['topic'], newer_timestamp) ### reset the consumer's position
                            raise exceptions.IncontiguousDataError(gap[0], gap[1], newer_timestamp) ### report the error

                        else:
                            pass ### there is a gap, but it doesn't touch anything we actually want so who cares?

                    timestamp = new_timestamp ### update the timestamp to be the new thing

                ### wait to poll for new data
                elapsed = timeit.default_timer() - start_time
                sleep_time = self.kwargs['retry_cadence'] - min(elapsed, self.kwargs['retry_cadence'])
                time.sleep(sleep_time)

            ### timestamp >= stop
            if self._data: ### if there was any data received, return that
                return self._data

            else: ### did not find ANY data within the requested time
                raise exceptions.NoDataError(self.start, self.end-self.start) ### did not find ANY data within the requested time

        else: ### if data was previously populated, return back the same data

            ### fill in missing channels
            if channels:
                for channel in channels:
                    if not self.is_cached(channel):
                        self._data[channel] = np.array([], dtype=self._dtype)

            return self._data

#-----------

class DruidClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for extracting triggers from Druid DataBases.
    """
    _default_columns = []
    _allowed_columns = None ### allow anything
    _required_kwargs = []

    def _retrieve_triggers(self, channel, segs, columns, verbose=False, **kwargs):
        """
        WARNING: not implemented!
        """
        raise NotImplementedError

#---------------------------------------------------------------------------------------------------
# objects used to report models, calibration_maps to disk locally or remotely
#---------------------------------------------------------------------------------------------------

class Reporter(object):
    """
    an object that handles I/O of models, calibration_maps
    this thing knows about file-systems and directory structures
    it could be extended to learn about database queries, etc

    the specifics about where things will be stored, etc should be set upon instantiation

    NOTE: this is just a parent class defining the API. Extensions of this class must define their own functionality

    Upon instantiation, we require some knowledge of where these things will be written (rootdir) as well as some knowledge of what time range the span (start, end)
    """
    _required_kwargs = []

    def __init__(self, rootdir, start, end, **kwargs):
        self._rootdir = rootdir

        self._start = start
        self._end = end

        for kwarg in self._required_kwargs:
            assert kwargs.has_key(kwarg), 'kwarg=%s required'%kwarg
        self.kwargs = kwargs

    @property
    def rootdir(self):
        return self._rootdir

    @rootdir.setter
    def rootdir(self, rootdir):
        self._rootdir = rootdir

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._end = end

    def preferred(self, nickname):
        return None

    def report(self, nickname, obj, preferred=False):
        """
        write a serialized version of the object into a path that's determined by self.kwargs
        """
        raise NotImplementedError

    def retrieve(self, nickname, preferred=False):
        """
        the inverse of report
        returns the object that was serialized

        if preferred, looks for the most recent object, which does not necessarily correspond to where this thing would write serialized output
        supported for convenient retrieval
        """
        raise NotImplementedError

    def glob(self, nickname, start, end):
        """return a generator over all existing data associated with nickname within [start, end)
        NOTE: this should work without modifying the object!
        """
        raise NotImplementedError

class DiskReporter(Reporter):
    """
    a sub-class of Reporter that reads/writes to the local disk (by pickling objects)
    this defines a directory structure that it expects and works within that paradigm

    NOTE:
        DiskReporter uses pickle serialization, which is fairly Python-specific and may not be particularily well compressed.
        There are extensions of this class that speak in more common data formats, like :class:idq.io.HDF5Reporter .
    """
    _suffix = 'txt'
    _cache_template = '.%s-preferred.cache' ### we'll always write plaintxt for this, so suffix doesn't really matter

    def __init__(self, *args, **kwargs):
        """
        delegates to Reporter.__init__ and then makes sure the correct directory exists
        """
        Reporter.__init__(self, *args, **kwargs)
        self._cache_mtime = dict() ### define this upon instantiation so each instance has it set independently

        if 'shmdir' in self.kwargs:
            shmdir = self.kwargs['shmdir']
            if not os.path.exists(shmdir):
                try:
                    os.makedirs(shmdir)
                except OSError: ### directory was created before this process could do it...
                    pass

    @property
    def suffix(self):
        return self._suffix

    def directory(self, start, end):
        """checks to make sure this directory exists whenever its called"""
        directory = names.start_end2dir(start, end, rootdir=names.start2dirMOD1e5(start, basename='START', rootdir=self.rootdir))
        if not os.path.exists(directory):
            try:
                os.makedirs(directory)
            except OSError: ### directory was created before this process could do it...
                pass
        return directory

    def path(self, nickname, start, end):
        """
        standardizes file naming conventions via delegations to idq.names
        """
        return os.path.join(
            self.directory(start, end),
            names.start_end2path(nickname, start, end, suffix=self.suffix)
        )

    def _cachepath(self, nickname):
        return os.path.join(
            self.rootdir,
            self._cache_template%(nickname),
        )

    def new_preferred(self, nickname):
        """
        determines if there is a new preferred path by checking the mtime of the cache
        Note: this updates self._cache_time, which is used to determine whether the cache has been touched since we last asked
        """
        cachepath = self._cachepath(nickname)
        if os.path.exists(cachepath): # path must exist for us to have a preferred
            mtime = os.path.getmtime(cachepath)
            if mtime != self._cache_mtime.get(nickname, None): ### check this against the stored mtime
                self._cache_mtime[nickname] = mtime # update the stored mtime
                return True
        return False

    def update_preferred(self, nickname, path):
        """
        write a path into the preferred cachefile
        NOTE: this is *not* automatically called within report. users must update the cache themselves.
        """
        cachepath = self._cachepath(nickname)
        with open(cachepath, 'a') as file_obj: ### FIXME: make this a truly atomic opperation!
            file_obj.write('\n'+path)

    def preferred(self, nickname):
        """
        get the path that redirects to the preferred file
        does this by referencing a cachepath that's automatically defined relative to rootdir
        """
        if not self.new_preferred(nickname):
            return None
        else:
            return self._tail(self._cachepath(nickname)) ### we can rely on this existing because new_preferred was True

    def _tail(self, path, byte_stride=80):
        result = []
        with open(path, 'r') as file_obj:
            file_obj.seek(0, 2) ### go to end of file
            tell = file_obj.tell()
            while tell > 0:
                stride = min(byte_stride, tell)

                file_obj.seek(-stride, 1) ### move back one stride relative to where we are
                line = file_obj.read(stride) ### read off some bytes and move the position within the file
                file_obj.seek(-stride, 1) ### go back further from where you are and then try again in next epoch

                if "\n" in line: # at least one line break in what we just read, apart from the one at the end
                    result.insert(0, line.strip().split('\n')[-1]) ### get only stuff on the last line
                    break

                else:
                    result.insert(0, line)
                    tell = file_obj.tell()

        return ''.join(result) if result else None

    def report(self, nickname, obj, preferred=False, **kwargs):
        """
        writes string into plaintxt file
        """
        path = self.path(nickname, self.start, self.end)
        self._write(path, obj, nickname=nickname, **kwargs)
        self._manage_shmdir(path, age=self.kwargs.get('shmdir_age', DEFAULT_SHMDIR_AGE))
        if preferred:
            self.update_preferred(nickname, path)
        return path

    def _write(self, path, string, **kwargs):
        with open(path, 'w') as file_obj: ### FIXME: make this a truly atomic operation
            file_obj.write(string)

    def retrieve(self, nickname, preferred=False):
        path = self.preferred(nickname) if preferred else self.path(nickname, self.start, self.end)
        if path is not None:
            return self._read(path)
        else:
            return None

    def _read(self, path):
        with open(path, 'r') as file_obj: ### FIXME: make this a truly atomic operation
            string = file_obj.read()
        return string

    def _manage_shmdir(self, path, age=DEFAULT_SHMDIR_AGE):
        if self.kwargs.has_key('shmdir'): ### assume this directory already exists, see __init__
            shmdir = self.kwargs['shmdir']
            shutil.copyfile(path, os.path.join(shmdir, os.path.basename(path)))
            self._cleanup(shmdir, age=age) ### get rid of old things in shmdir

    def _cleanup(self, directory, age=DEFAULT_SHMDIR_AGE):
        """
        Delete files in directory that are older than age given
        """
        for file_ in os.listdir(directory):
            filepath = os.path.join(directory, file_)
            try:
                if time.time() - os.path.getmtime(filepath) > age: ### file hasn't been modified in a while
                    os.remove(filepath)
            except OSError: ### file was already deleted, can happen with parallel workflows
                pass

    def glob_paths(self, nickname, start, end):
        """a generator for the set of paths needed by glob
        NOTE: if there are overlapping ranges, all will be returned. They are sorted by start time and ties are broken by duration (shorter durations first)
        ALSO NOTE: this returns all data with any overlap with the requested range and does NOT truncate or extend it to fit the specified range.
        """
        bounds = []
        for directory in sorted(glob.glob(os.path.join(self.rootdir, '*'))):
            if os.path.isdir(directory):
                try:
                    s, e = names.dir2start_end(directory)
                    if s > end: ### we've run off the end
                        break # so truncate the iteration
                    # we already checked (s<end), so we don't check it again
                    if utils.livetime(utils.segments_intersection([[s,e]], [[start,end]])): ### some overlap
                        bounds.append((s, e))
                except ValueError: ### bad formatting
                    pass

        bounds.sort(key=lambda _: _[1]) # sort by end time first
        bounds.sort(key=lambda _: _[0]) # then sort by start time
        for start, end in bounds:
            data = None
            path = self.path(nickname, start, end)
            if os.path.exists(path): ### there is data for this nickname in this directory
                yield start, end, path

    def glob(self, nickname, start, end):
        """return a generator over all data associated with nickname within [start, end)
        """
        for s, e, path in self.glob_paths(nickname, start, end):
            data = self._read(path)
            yield s, e, data

class PickleReporter(DiskReporter):
    """
    pickle objects instead of just writing strings into plaintxt
    """
    _suffix = 'pkl'

    def _write(self, path, obj, **kwargs):
        """
        pickles obj into path
        """
        with open(path, 'w') as file_obj: ### FIXME: make this a truly atomic operation
            pickle.dump(obj, file_obj)

    def _read(self, path):
        """
        unpickles obj from self.path(nickname)
        """
        with open(path, 'r') as file_obj: ### FIXME: make this a truly atomic operation
            obj = pickle.load(file_obj)
        return obj

class HDF5Reporter(DiskReporter):
    """
    a variant of DiskReporter that expects objects to fit into a tabular data format and writes that to disk as hdf5
    """
    _suffix = 'hdf5'

    def _write(self, path, data, **attrs):
        with h5py.File(path, 'w') as file_obj: ### FIXME: make this a truly atomic operation
            dataset = file_obj.create_dataset('data', data=data)
            for key, value in attrs.items():
                dataset.attrs.create(key, value)

    def _read(self, path):
        with h5py.File(path, 'r') as file_obj: ### FIXME: make this a truly atomic operation
            data = file_obj['data'][...]
            attrs = dict(file_obj['data'].attrs.items())
        return (data, attrs)

class SegmentReporter(DiskReporter):
    """
    a variant of DiskReporter that expects to only deal with ligolw segments in XML form.
    """
    _suffix = 'xml.gz'

    def __init__(self, *args, **kwargs):
        DiskReporter.__init__(self, *args, **kwargs)
        self.instrument = self.kwargs.pop('instrument') if self.kwargs.has_key('instrument') else 'None'

    def report(self, nickname, obj, preferred=False, **kwargs):
        path = self.path(nickname, self.start, self.end)
        self._write(path, obj, nickname=nickname, **kwargs)
        self._manage_shmdir(path, age=self.kwargs.get('shmdir_age', DEFAULT_SHMDIR_AGE))
        if preferred:
            self.update_preferred(nickname, path)
        return path

    def _write(self, path, segs, nickname='name', **kwargs):
        # create xmldoc and process
        xmldoc = ligolw.Document()
        xmldoc.appendChild(ligolw.LIGO_LW())
        process = lsctables.Process(process_id = lsctables.ProcessTable.get_next_id())

        # store segments into xmldoc
        with ligolw_segments.LigolwSegments(xmldoc, process) as xmlsegments:
            xmlsegments.insert_from_segmentlistdict({self.instrument: segs}, nickname)

        # save to disk
        ligolw_utils.write_filename(xmldoc, path, gz=path.endswith('gz'))

    def _read(self, path):
        segs = ligolw_segments.segmenttable_get_by_name(ligolw_utils.load_filename(path, contenthandler=ligolw_segments.LIGOLWContentHandler), path).coalesce()
        return segs[self.instrument]

class HDF5SegmentReporter(HDF5Reporter):
    """
    a variant of HDF5 reporter that expects to only deal with segment lists

    NOTE:
        we may need to re-examine this and use a different format than HDF5.
        for example, JSON or ligolw xml may be necessary so that others can read these in trivially.
        that's probably a not-uncommon use case, so it definitely deserves some thought
    """

    _dtypes = [('start_sec', '>i4'), ('start_ns', '>i4'), ('end_sec', '>i4'), ('end_ns', '>i4')]

    def _write(self, path, segments, **kwargs):
        segments = np.array(segments)
        seg_start, seg_start_ns = utils.float2sec_ns(segments[:,0])
        seg_end, seg_end_ns = utils.float2sec_ns(segments[:,1])
        data = np.array([seg_start, seg_start_ns, seg_end, seg_end_ns], dtype=self._dtypes)

        start_sec, end_ns = utils.float2sec_ns(self.start)
        end_sec, end_ns = utils.float2sec_ns(self.end)

        return HDF5Reporter._write(self, path, data, start_sec=start_sec, start_ns=start_ns, end_sec=end_sec, end_ns=end_ns, **kwargs)

    def _read(self, path):
        data, attrs = HDF5Reporter._read(self, path)
        start = utils.sec_ns2float(data['start_sec'], data['start_ns'])
        end = utils.sec_ns2float(data['end_sec'], data['end_ns'])
        return np.transpose(np.array([start, end]))

class GPSTimesReporter(HDF5Reporter):
    """
    a variant of HDF5 reporter that expects to only deal with lists of gps times
    """

    _dtypes = [('gps_sec', '>i4'), ('gps_ns', '>i4')]

    def _write(self, path, times, **kwargs):
        data = np.array(times, dtype=self._dtypes)
        start_sec, start_ns = utils.float2sec_ns(self.start)
        end_sec, end_ns = utils.float2sec_ns(self.end)
        return HDF5Reporter._write(self, path, data, start_sec=start_sec, start_ns=start_ns, end_sec=end_sec, end_ns=end_ns, **kwargs)

    def _read(self, path):
        data, attrs = HDF5Reporter._read(self, path)
        return (utils.sec_ns2float(data['gps_sec'], data['gps_ns']), utils.sec_ns2float(attrs['start_sec'], attrs['start_ns']), utils.sec_ns2float(attrs['end_sec'], attrs['end_ns']))

QUIVER_HDF5_DTYPE = [('gps', 'float'), ('rank','float'), ('label', 'float'), ('hash', 'S21')]
QUIVER_HDF5_NAME = 'quiver'
def _quiver2dataset(h5group, quiver):
    """a helper function to allow us to re-use quiver <--> hdf5 logic
    """
    data = []
    for vector in quiver:
        data.append((vector.gps, vector.rank, vector.label, str(vector.hash)))
    dataset = h5group.create_dataset(QUIVER_HDF5_NAME, data=np.array(data, dtype=QUIVER_HDF5_DTYPE))

QUIVER_HDF5_SEG_DTYPE = [('start_sec', '>i4'), ('start_ns', '>i4'), ('end_sec', '>i4'), ('end_ns', '>i4')]
QUIVER_HDF5_SEG_NAME = 'quiver_segments'
def _quiver_segs2dataset(h5group, start, end, segs):
    """a helper function to allow us to re-use quiver <--> hdf5 logic
    """
    h5segs = []
    for s, e in segs:
        start_sec, start_ns = utils.float2sec_ns(s)
        end_sec, end_ns = utils.float2sec_ns(e)
        h5segs.append((start_sec, start_ns, end_sec, end_ns))
    dataset = h5group.create_dataset(QUIVER_HDF5_SEG_NAME, data=np.array(h5segs, dtype=QUIVER_HDF5_SEG_DTYPE))

    ### attrs
    start_sec, start_ns = utils.float2sec_ns(start)
    end_sec, end_ns = utils.float2sec_ns(end)
    for key, value in [('start_sec',start_sec), ('start_ns',start_ns), ('end_sec',end_sec), ('end_ns',end_ns)]:
        dataset.attrs.create(key, value)

def _h5group2quiver(h5group):
    """a helper function to allow us to re-use quiver <--> hdf5 logic
    """
    quiver = features.Quiver()
    if QUIVER_HDF5_SEG_NAME in h5group.keys(): ### extract segment information
        attrs = dict(h5group[QUIVER_HDF5_SEG_NAME].attrs.items())
        quiver.start = utils.sec_ns2float(attrs['start_sec'], attrs['start_ns'])
        quiver.end = utils.sec_ns2float(attrs['end_sec'], attrs['end_ns'])

        quiver.segs = ligo_segments.segmentlist([])
        for start_sec, start_ns, end_sec, end_ns in h5group[QUIVER_HDF5_SEG_NAME][...]:
            quiver.segs.append(ligo_segments.segment(utils.sec_ns2float(start_sec, start_ns), utils.sec_ns2float(end_sec, end_ns)))

    for row in h5group[QUIVER_HDF5_NAME][...]: ### add in the FeatureVector data
        vector = features.FeatureVector(
            row['gps'],
            row['label'],
            None, ### we don't store classifier_data reference...
        )
        vector.rank = None if row['rank']!=row['rank'] else row['rank']
        vector.hash = None if row['hash']=='None' else row['hash']

        quiver.append(vector)

    return quiver

class QuiverReporter(HDF5Reporter):
    """
    store high-level attributes from a quiver
    """
    def _nickname2segsname(self, nickname):
        return nickname+'_segments'

    def _write(self, path, quiver, **kwargs):
        ### ranks and all that crap
        with h5py.File(path, 'w') as file_obj: ### FIXME: make this a truly atomic operation
            _quiver2dataset(file_obj, quiver)
            if quiver.segs is not None:
                _quiver_segs2dataset(file_obj, quiver.start, quiver.end, quiver.segs)

    def _read(self, path):
        with h5py.File(path, 'r') as file_obj: ### FIXME: make this a truly atomic operation
            quiver = _h5group2quiver(file_obj)
        return quiver

class CalibrationMapReporter(HDF5Reporter):
    """
    store high-level information from a calibration map
    based on HDF5Reporter but extends this to use a more complicated structure under the hood
    """
    _samples_dtype = [('sample', 'float')]
    _interp_dtype = [
        ('x', 'float'),
        ('logpdf', 'float'),
        ('logpdf_m1', 'float'),
        ('logpdf_m2', 'float'),
        ('cdf', 'float'),
        ('cdf_m1', 'float'),
        ('cdf_m2', 'float'),
        ('pdf_alpha', 'float'),
        ('pdf_beta', 'float'),
        ('cdf_alpha', 'float'),
        ('cdf_beta', 'float'),
    ]

    def _write(self, path, calibrationmap, **kwargs):
        with h5py.File(path, 'w') as file_obj: ### FIXME: make this a truly atomic operation
            group = file_obj.create_group('CalibrationMap')

            # store loglike_cdf interpolation object
            dataset = group.create_dataset('interp_loglike_cdf', data=calibrationmap._interp_loglike_cdf)
            dataset.attrs.create('num_points', len(calibrationmap._ranks))
            dataset.attrs.create('num_mcmc', int(round(1./calibrationmap._dq, 0)))

            # store individual interpolation objects
            for name, quiver, kde in [('gch', calibrationmap._gch_quiver, calibrationmap._gch_kde), ('cln', calibrationmap._cln_quiver, calibrationmap._cln_kde)]:
                g = group.create_group(name)

                ### store the quivers
                _quiver2dataset(g, quiver)
                if quiver.segs is not None:
                    _quiver_segs2dataset(g, quiver.start, quiver.end, quiver.segs)

                ### store the processed KDES
                data = np.array(
                    zip(
                        kde._interp_x,
                        kde._interp_logpdf,
                        kde._interp_logpdf_m1[:,0],
                        kde._interp_logpdf_m2[:,0],
                        kde._interp_cdf,
                        kde._interp_cdf_m1[:,0],
                        kde._interp_cdf_m2[:,0],
                        kde._interp_pdf_alpha,
                        kde._interp_pdf_beta,
                        kde._interp_cdf_alpha,
                        kde._interp_cdf_beta,
                    ),
                    dtype=self._interp_dtype
                )
                dataset = g.create_dataset('interp', data=data)
                dataset.attrs.create('b', kde.b)
                dataset.attrs.create('num_points', kde._num_points)
                dataset.attrs.create('count', kde._count)

    def _read(self, path):
        with h5py.File(path, 'r') as file_obj: ### FIXME: make this a truly atomic operation
            group = file_obj['CalibrationMap']
            interp_loglike = group['interp_loglike_cdf']

            ### pull out quivers
            gch = _h5group2quiver(group['gch'])
            cln = _h5group2quiver(group['cln'])

            ### pull out interpolated things
            gch_interp = group['gch/interp']
            cln_interp = group['cln/interp']

            ### set up the calibration map's provenance
            dummy_quiver = features.Quiver()
            dummy_quiver.start = min(gch.start, cln.start)
            dummy_quiver.end = max(gch.end, cln.end)
            dummy_quiver.segs = utils.segments_union(gch.segs, cln.segs)
            calibrationmap = calibration.CalibrationMap(
                dummy_quiver,
                num_mcmc=interp_loglike.attrs['num_mcmc'],
                num_points=interp_loglike.attrs['num_points'],
                gch_num_points=gch_interp.attrs['num_points'],
                gch_b=gch_interp.attrs['b'],
                cln_num_points=cln_interp.attrs['num_points'],
                cln_b=cln_interp.attrs['b'],
                compute=False,
            )

            ### fill in quivers
            calibrationmap._gch_quiver = gch
            calibrationmap._cln_quiver = cln

            ### fill in interpolation objects for loglike_quantiles
            calibrationmap._interp_loglike_cdf = interp_loglike[...] ### just copy the answer straight out of the hdf5 file

            ### fill in recorded interpolation objects
            for kde, quiver, obj in [(calibrationmap._gch_kde, gch, gch_interp), (calibrationmap._cln_kde, cln, cln_interp)]:
                ### high-level stuff
                kde._obs = quiver.ranks
                kde._count = obj.attrs['count']

                ### interpolation stuff
                kde._interp_x[:] = obj['x']
                kde._interp_logpdf[:] = obj['logpdf']
                kde._interp_logpdf_m1[:,0] = obj['logpdf_m1']
                kde._interp_logpdf_m2[:,0] = obj['logpdf_m2']
                kde._interp_cdf = obj['cdf']
                kde._interp_cdf_m1[:,0] = obj['cdf_m1']
                kde._interp_cdf_m2[:,0] = obj['cdf_m2']
                kde._interp_pdf_alpha[:] = obj['pdf_alpha']
                kde._interp_pdf_beta[:] = obj['pdf_beta']
                kde._interp_cdf_alpha[:] = obj['cdf_alpha']
                kde._interp_cdf_beta[:] = obj['cdf_beta']

        calibrationmap.compute_rates() ### compute the rates used for the prior odds and such
        return calibrationmap

HDF5SERIES_REPORTER_DATASET_TEMPLATE = '%d_%d'
class HDF5SeriesReporter(HDF5Reporter):
    """
    store series information in hdf5 files
    """
    _T0_NAME = 't0'
    _DELTAT_NAME = 'deltaT'
    _MODEL_NAME = 'idq_model'
    _CALIB_NAME = 'idq_calib'

    def _write(self, path, seriesdict, **kwargs):
        if isinstance(seriesdict, dict):
            seriesdict = [seriesdict]

        with h5py.File(path, 'w') as file_obj: ### FIXME: make this a truly atomic operation
            group = file_obj.create_group('series')

            for consecutive_data in seriesdict:
                dataset_name = HDF5SERIES_REPORTER_DATASET_TEMPLATE%(int(consecutive_data['t0']), int(consecutive_data['deltaT']))
                keys = consecutive_data['data'].keys()

                data = np.array(zip(*(consecutive_data['data'][key] for key in keys)), dtype=[(key, 'float') for key in keys])

                dataset = group.create_dataset(dataset_name, data=data)
                dataset.attrs.create(self._T0_NAME, consecutive_data['t0'])
                dataset.attrs.create(self._DELTAT_NAME, consecutive_data['deltaT'])
                dataset.attrs.create(self._MODEL_NAME, consecutive_data['model']) ### record the associated model hash
                dataset.attrs.create(self._CALIB_NAME, consecutive_data['calib']) ### and the calibration map hash

                for key in keys:
                    dataset.attrs.create(key+'_unit', '')

    def _read(self, path):
        seriesdict = []
        with h5py.File(path, 'r') as file_obj:
            group = file_obj['series']
            for key in group.keys():
                dataset = group[key]
                d = dict()
                d['t0'] = dataset.attrs[self._T0_NAME]
                d['deltaT'] = dataset.attrs[self._DELTAT_NAME]
                d['model'] = dataset.attrs[self._MODEL_NAME]
                d['calib'] = dataset.attrs[self._CALIB_NAME]

                d['data'] = dict()
                for field in dataset[...].dtype.fields.keys():
                    d['data'][field] = dataset[field]

                seriesdict.append(d)

        return seriesdict

class GWFSeriesReporter(DiskReporter):
    """
    store series informaiton in gwf files

    NOTE:

      * GWFSeriesReporter currently only supports a single contiguous segment and will raise an exception if it finds more than one.
      * This is most likely the most common use case (e.g.: streaming timeseries ignoring segdb), so it's probably fine
    """
    _suffix = 'gwf'

    _MODEL_NAME = 'IDQ_MODEL'
    _CALIB_NAME = 'IDQ_CALIB'

    def _write(self, path, seriesdict, **kwargs):
        if isinstance(seriesdict, list):
            assert len(seriesdict)==1, 'GWFSeriesReporter currently only supports a single contiguous segment of data'
            seriesdict = seriesdict[0]

        ### make the frame
        frame = frcpp.FrameH()

        ### add provenance information
        frame.AppendFrHistory(frcpp.FrHistory(self._MODEL_NAME, utils.current_gps_time(), seriesdict['model']))
        frame.AppendFrHistory(frcpp.FrHistory(self._CALIB_NAME, utils.current_gps_time(), seriesdict['calib']))

        ### add data
        t0 = seriesdict['t0']
        deltaT = seriesdict['deltaT']
        for key, data in seriesdict['data'].items():
            N = len(data)
            # make a vector
            frvect = frcpp.FrVect(
                key,
                frcpp.FrVect.FR_VECT_8R,
                1, ### number of dimensions
                frcpp.Dimension( # the dimension object
                    N, ### number of elements
                    deltaT, ### spacing
                    's', ### units
                    t0, ### start time
                ),
                '', ### dimensionless unit, cause we have probabilities and such
            )
            frvect.GetDataArray()[:] = data ### actually populate the object

            # add that vector to the ProcData
            frdata = frcpp.FrProcData(
                key,
                '', ### no comment
                frcpp.FrProcData.TIME_SERIES,  # ID as time-series
                frcpp.FrProcData.UNKNOWN_SUB_TYPE,  # empty sub-type (fseries)
                0,  # offset of first sample relative to frame start
                N*deltaT,  # duration of data
                0.,  # heterodyne frequency
                0.,  # phase of heterodyne
                0.,  # frequency range
                0.,  # resolution bandwidth
            )
            frdata.AppendData(frvect) ### add the vector to this data

            # add ProcData to FrameH
            frame.AppendFrProcData(frdata)

        frame.SetDt(deltaT*N) ### set duration

        ### write
        frame.Write(frcpp.OFrameFStream(path))

    def _read(self, path):
        seriesdict = dict()
        stream = frcpp.IFrameFStream(path)
        frame = stream.ReadNextFrame() ### assume a single frame

        ### extract data
        toc = stream.GetTOC()
        seriesdict = dict()
        for channels, foo in [(toc.GetProc(), stream.ReadFrProcData), (toc.GetADC(), stream.ReadFrAdcData), (toc.GetSim(), stream.ReadFrSimData)]:
            for channel in channels:
                data = foo(0, channel) ### magic number: one frame per stream...
                dim = data.data[0].GetDim(0)
                data = data.data[0].GetDataArray() ### magic number: one FrVect wrapper per channel
                seriesdict[channel] = np.empty_like(data, dtype=float)
                seriesdict[channel][:] = data[:]

        # assume there is at least one channel
        ### NOTE: these should be the same for all channels, so we only read them once
        t0 = dim.startX
        deltaT = dim.dx

        ### extract histories
        histories = dict()
        for history in frame.RefHistory():
            histories[history.GetName()] = history.GetComment()

        for name in [self._MODEL_NAME, self._CALIB_NAME]:
            if name not in histories:
                raise KeyError('could not find FrHistory named=%s in %s!'%(name, path))

        seriesdict = {
            'data':seriesdict,
            't0':t0,
            'deltaT':deltaT,
            'model':histories[self._MODEL_NAME],
            'calib':histories[self._CALIB_NAME],
        }

        return [seriesdict] ### we only know how to handle a single contiguous stretch of data at the moment

class JSONMetricReporter(DiskReporter):
    """
    Store metrics and dimensions in JSON-compatible rows. Used mainly for plotting within idq-report.

    NOTE: It is assumed that data is stored in the following form:
        {
            'dimension_names': [dimension_name1, ..., dimension_nameM],
            'metric_names': [metric_name1, ..., metric_nameN],
            'data':{(dimension1, ..., dimensionM): [iterable(metric1), ..., iterable(metricN)]}
        }

        where each iterable is of the same length.
    """
    _suffix = 'json'

    def _write(self, path, jsondict, **kwargs):
        dimension_names = jsondict['dimension_names']
        metric_names = jsondict['metric_names']
        out_rows = []

        # create JSON-formatted rows
        for dimension_keys, metrics in jsondict['data'].items():
            dimension_row = {dimension_name: dimension for dimension_name, dimension in zip(dimension_names, dimension_keys)}
            for row in zip(*metrics):
               metric_row = {metric_name: metric for metric_name, metric in zip(metric_names, row)}
               out_row = dict(dimension_row)
               out_row.update(metric_row)
               out_rows.append(out_row)

        # format output JSON
        outdict = dict(jsondict)
        outdict['data'] = out_rows

        # save to disk
        with open(path, 'w') as f:
            json.dump(outdict, f)

    def _read(self, path):
        # load from disk
        indict = {}
        with open(path, 'r') as f:
            indict = json.load(f)

        # format input JSON
        jsondict = dict(indict)
        dimension_names = indict['dimension_names']
        metric_names = indict['metric_names']
        jsondict['data'] = defaultdict(lambda: [[] for metric in metric_names])

        for row in indict['data']:
            dim_key = tuple(row[dimension] for dimension in dimension_names)
            for idx, metric in enumerate(metric_names):
                jsondict['data'][dim_key][idx].append(row[metric])

        return jsondict

#---------------------------------------------------------------------------------------------------
# stream-based reporters used to stream feature vectors, models, and maps via various protocols
#---------------------------------------------------------------------------------------------------

class StreamReporter(Reporter):
    """
    a sub-class of Reporter that retrieves data (feature vectors, models, maps)
    using various streaming protocols

    this object knows how to retrieve/send data in various
    fashions (synchronous or asynchronous) given all relevant configuration options
    and a route (rootdir) where data transfer takes place

    can push relevant data to this object via report() and pull using retrieve()

    NOTE: this is just a parent class defining the API. Extensions of this class must define their own functionality

    Upon instantiation, we require a route (rootdir) as well as custom configuration options specified by keyword args
    """
    def __init__(self, *args, **kwargs):
        self._timeout = kwargs.pop('timeout', DEFAULT_STREAMREPORTER_TIMEOUT)
        Reporter.__init__(self, *args, **kwargs)

    @property
    def timeout(self):
        return self._timeout

class DruidReporter(StreamReporter):
    """
    a sub-class of StreamReporter that reads/writes to an instance of Druid
    defines URL conventions for where things will be located within Druid, etc
    """

    def report(self, nickname, obj, preffered=False):
        raise NotImplementedError

    def retrieve(self, nickname, preferred=False):
        raise NotImplementedError

class HTTPReporter(StreamReporter):
    """
    a sub-class of StreamReporter that reads/writes through Bottle Routes or other HTTP-based services
    """

    def report(self, nickname, obj, preferred=False):
        raise NotImplementedError

    def retrieve(self, nickname, preferred=False):
        raise NotImplementedError

class KafkaReporter(StreamReporter):
    """
    a sub-class of StreamReporter that reads/writes through Kafka topics.

    The rootdir argument is used as a proxy for the Kafka server that topics are hosted on,
    which is the kafka setting 'bootstrap.servers'. In addition, the 'group.id' kwarg needs to be
    specified upon instantiation. Any other kwargs can be passed in and will be used for the Kafka
    configuration to create producers and consumers at will.

    This Reporter needs to be able to connect to a Kafka instance for it to work properly.
    """
    _required_kwargs = ['group', 'port']

    def __init__(self, *args, **kwargs):
        utils.check_kafka()

        StreamReporter.__init__(self, *args, **kwargs)
        self._old = False

        # set up container for producers/consumers
        self._kafka_kwargs = self.kwargs.get('kafka_kwargs', {})
        self.consumers = {}
        self.producers = {}

    @property
    def start(self):
        """it appears I have to re-define this because inheritance doesn't take care of this for me"""
        return self._start

    @start.setter
    def start(self, start):
        t = self.timestamp ### current timestamp
        self._start = start ### update
        self._old = self.timestamp <= t ### figure out whether we're at an older timestamp than we were

    @property
    def end(self):
        """it appears I have to re-define this because inheritance doesn't take care of this for me"""
        return self._end

    @end.setter
    def end(self, end):
        t = self.timestamp ### current timestamp
        self._end = end ### update
        self._old = self.timestamp <= t ### figure out whether we're at an older timestamp than we were

    @property
    def timestamp(self):
        return 0.5*(self.start+self.end)

    @property
    def old(self):
        return self._old

    def _register(self, process_type, nickname):
        if process_type == 'producer':
            self._register_producer(nickname)
        elif process_type == 'consumer':
            self._register_consumer(nickname)
        else:
            raise ValueError, "Process type can only be 'consumer' or 'producer'"

    def _register_producer(self, nickname):
        self.producers[nickname] = utils.KafkaProducer(self.kwargs['group'], self.kwargs['port'], **self._kafka_kwargs)

    def _register_consumer(self, nickname):
        self.consumers[nickname] = utils.KafkaConsumer(self.kwargs['group'], self.kwargs['port'], topics=[nickname], **self._kafka_kwargs)

    def report(self, nickname, value, preferred=False):
        """
        Creates a Kafka producer to send messages (value) to a given Kafka topic (nickname),
        with a given timestamp (timestamp).

        Uses pickle to serialize messages.
        """
        if self.old:
            raise ValueError('KafkaReporter has already reported a timestamp larger than %.3f'%self.timestamp)

        ### create producer if one not already allocated
        if nickname not in self.producers:
            self._register_producer(nickname)

        if isinstance(value, features.Quiver): ### don't transfer classifier_data reference for quiver
            for vector in value:
                vector.classifier_data = None

        ### push data to the given topic (nickname)
        ### NOTE: we take no action based on preferred here because that's handled by self.old
        self.producers[nickname].produce(timestamp=self.timestamp, topic=nickname, value=pickle.dumps(value))

        ### flush out topic of sent packets (block until message is sent)
        self.producers[nickname].flush()

        return 'kafka topic= %s from %s'%(nickname, self.kwargs['port'])

    def retrieve(self, nickname, preferred=False):
        """
        Creates a Kafka consumer to retrieve messages from a given Kafka topic (nickname).

        Uses pickle to deserialize messages.
        """
        ### create consumer if one not already allocated
        if nickname not in self.consumers:
            self._register_consumer(nickname)

            ### FIXME: why do I need to get a message first before I can seek, even when assigning to a particular partition?
            message = None
            while not message:
                message = self.consumers[nickname].poll(timeout=self.timeout)

            ### point consumer to latest buffer in topic
            self.consumers[nickname].seek_to_latest(nickname)

        if preferred: ### just consume the next buffer in the topic
            message = self.consumers[nickname].poll(timeout=self.timeout)
        else:
            self.consumers[nickname].seek_timestamp(nickname, self.timestamp) ### go to offset for this timestamp
            message = self.consumers[nickname].poll(timeout=self.timeout) ### then consume

        if not message:
            return None

        elif message.error():
            if message.error().code() == utils.kafka.KafkaError._PARTITION_EOF:
                return None
            else:
                raise message.error()

        else:
            return pickle.loads(message.value())
