__description__ = "a module that houses methods and classes used to produce html reports"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import os
import json
from collections import defaultdict
import numpy as np

### dependencies not in standard Python libraries
try:
    import jinja2 ### for generating html pages
except ImportError:
    jinja2 = None

from idq import configparser
from idq import plots
from idq import names
from idq import features
from idq import factories
from idq import logs
from idq import utils

from ligo import segments as ligo_segments

from idq import __version__ ### import this for the "entire library"
__process_name__ = "iDQ"

#-------------------------------------------------

DEFAULT_FIGTYPE = 'png'

DEFAULT_GRACEID = None
try:
    from dqr import json as dqr_json
    from dqr import logging as dqr_logging
except ImportError:
    dqr_json = dqr_logging = DEFAULT_GRACEDB_URL = None

try:
    from ligo.gracedb.rest import GraceDb, DEFAULT_SERVICE_URL as DEFAULT_GRACEDB_URL
except ImportError:
    GraceDb = DEFAULT_GRACEDB_URL = None

#-------------------------------------------------
### reporting classes/utilities

class HTMLReport(object):
    """
    Generates HTML-based reports for summarizing results and producing visualizations.
    """
    def __init__(self, config):
        self.config = config
        self.environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(os.path.join(os.environ['IDQ_PATH'], 'web/templates')),
            autoescape=jinja2.select_autoescape(['html'])
        )
        ### FIXME: this would be the preferred method once this works correctly
        #self.environment = jinja2.Environment(
        #    loader=jinja2.PackageLoader('idq', 'web/templates'),
        #    autoescape=jinja2.select_autoescape(['html'])
        #)

        ### base settings
        self.base_settings = {
            'rootdir': os.path.abspath(os.path.realpath(config.get('general','rootdir'))),
            'instrument': config.get('general', 'instrument'),
            'tag': config.get('general', 'tag')
        }

        ### classifier settings
        classifier_names = config.get('general', 'classifiers').split()
        classifiers = {name: configparser.evalkwargs(**dict(config.items(name))) for name in classifier_names}
        self.classifier_settings = {
            'classifier_names': classifier_names,
            'classifiers': classifiers
        }

        ### sampling settings
        self.sampling_settings = {
            'target_channel': config.get('samples', 'target_channel'),
            'target_bounds': configparser.config2bounds(config.get('samples', 'target_bounds')),
            'target_channel': configparser.config2bounds(config.get('samples', 'dirty_bounds')),
            'target_channel': config.getfloat('samples', 'dirty_window'),
            'sample_rate_train': config.getfloat('train', 'random_rate'),
            'sample_rate_evaluate': config.getfloat('evaluate', 'random_rate'),
            'sample_rate_timeseries': config.getfloat('timeseries', 'srate')
        }

        ### web settings
        self.web_settings = {
            'webdir': config.get('general', 'webdir') if config.has_option('general', 'webdir') else rootdir,
        }

        ### define pages that are available
        self.pages = ['Summary']
        # FIXME: these are rough sketch of the intended pages, but need to be implemented
        #self.pages = ['Summary', 'Configuration', 'Features', 'Classifiers']

        ### create webdir if it doesn't exist
        if not os.path.exists(self.web_settings['webdir']):
            os.makedirs(self.web_settings['webdir'])

    def generate_summary(self, plots):
        """
        Creates a summary page of important results.
        """
        template = self.environment.get_template('summary.html')

        ### create kwargs to pass to template to render
        template_kwargs = {
            'current': 'Summary',
            'tabs': self.pages,
            'plot_results': plots
        }

        ### write summary html file
        webpath = os.path.join(self.web_settings['webdir'], 'index.html')
        with open(webpath, 'w') as f:
            f.write(template.render(**template_kwargs))

    def generate_config(self, **kwargs):
        """
        Creates a page summarizing important configuration settings.
        """
        raise NotImplementedError

    def generate_features(self, **kwargs):
        """
        Creates a page summarizing features used.
        """
        raise NotImplementedError

    def generate_classifiers(self, **kwargs):
        """
        Creates a page summarizing results from classifiers used.
        """
        raise NotImplementedError

#-------------------------------------------------

class DQReport(object):
    """generate everything needed for a DQR report
    """
    _title_fontsize = 14

    def __init__(self, config_path, start, end, t0=None, zoom_start=None, zoom_end=None):
        self._start = start
        self._end = end
        self._t0 = t0
        self._zoom_start = zoom_start
        self._zoom_end = zoom_end
        self._path2config(config_path)

    @property
    def config(self):
        return self._config

    @property
    def config_path(self):
        return self._config_path

    @property
    def instrument(self):
        return self._instrument

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def zoom_start(self):
        return self._zoom_start if (self._zoom_start is not None) else self._start

    @property
    def zoom_end(self):
        return self._zoom_end if (self._zoom_end is not None) else self._end

    @property
    def t0(self):
        return self._t0 if (self._t0 is not None) else 0.5*(self.zoom_start+self.zoom_end)

    @property
    def tag(self):
        return self._tag

    @property
    def classifier_data(self):
        return self._classifier_data

    @property
    def quiver_factory(self):
        return factories.QuiverFactory(self.classifier_data)

    def _path2config(self, config_path):
        self._config_path = os.path.abspath(os.path.realpath(config_path))
        self._config = configparser.path2config(config_path)
        self._instrument = self.config.get('general', 'instrument', '')
        self._tag = self.config.get('general', 'tag')

        rootdir = os.path.abspath(os.path.realpath(self.config.get('general','rootdir')))
        timeseriesdir = names.tag2timeseriesdir(self.tag, rootdir=rootdir)
        traindir = names.tag2traindir(self.tag, rootdir=rootdir)
        evaluatedir = names.tag2evaluatedir(self.tag, rootdir=rootdir)
        calibratedir = names.tag2calibratedir(self.tag, rootdir=rootdir)

        ### reporters
        reporter_factory = factories.ReporterFactory()
        # timeseries
        ritems = configparser.evalkwargs(**dict(self.config.items('timeseries reporting')))
        self.timeseriesreporter = reporter_factory(timeseriesdir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        # train
        ritems = configparser.evalkwargs(**dict(self.config.items('train reporting')))
        self.trainreporter = reporter_factory(traindir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        # evaluate
        ritems = configparser.evalkwargs(**dict(self.config.items('evaluate reporting')))
        self.evaluatereporter = reporter_factory(evaluatedir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        # calibrate
        ritems = configparser.evalkwargs(**dict(self.config.items('calibrate reporting')))
        self.calibratereporter = reporter_factory(calibratedir, self.start, self.end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(self.tag, 'timeseries')))

        ### classifier data
        classifier_data_factory = factories.ClassifierDataFactory()
        citems = configparser.evalkwargs(**dict(self.config.items('timeseries data discovery')))
        self._classifier_data = classifier_data_factory(self.zoom_start, self.zoom_end, **configparser.add_missing_kwargs(citems, group=names.tag2group(self.tag, 'timeseries')))

    def report(self,
            nicknames,
            graceid=DEFAULT_GRACEID,
            gracedb_url=DEFAULT_GRACEDB_URL,
            output_dir='.',
            verbose=False,
            figtype=DEFAULT_FIGTYPE,
            annotate_gch_gps=False,
            links=[],
            pglitch_thr=0.9,
        ):
        """actually report everything for nickname
        """
        if dqr_json is None:
            raise ImportError('dqr libraries not available!')

        title_graceid = 'for %s '%graced if (graceid is not None) else ''

        output_dir = os.path.join(output_dir, '%s%s%s-%s'%(self.instrument, __process_name__, self.tag, graceid))
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        logger = logs.get_logger(
            names.tag2dqr_logname(self.tag+'-'+self.instrument, graceid),
            rootdir=output_dir,
            verbose=verbose
        )
        logger.info('generating DQReport for %s within [%.3f, %.3f) with %s'%(', '.join(nicknames), self.start, self.end, self.config_path))

        # storate objects
        uploads = []
        figures = []
        tables = []

        ### go get all the data you'll need
        # series
        logger.info('globbing timeseries within [%.3f, %.3f)'%(self.start, self.end))
        series = defaultdict(list)
        for nickname in nicknames:
            for _, _, S in self.timeseriesreporter.glob(names.nickname2timeseries_topic(self.instrument, nickname), self.start, self.end):
                for s in S:
                    start = s['t0']
                    dt = s['deltaT']
                    end = start + dt*len(s['data'].values()[0])
                    times = np.arange(start, end, dt)
                    s['data']['times'] = times

                    if start < self.start: ### truncate the start
                        truth = s['data']['times']>=self.start
                        for key, val in s['data'].items():
                            s['data'][key] = val[truth]
                        s['t0'] = self.start

                    if end > self.end: ### turncate the end
                        truth = s['data']['times']<self.end
                        for key, val in s['data'].items():
                            s['data'][key] = val[truth]

                    series[nickname].append(s)

        # calibration maps and models
        logger.info('retrieving CalibrationMaps and Models corresponding to %d series'%len(series))
        calibmaps = dict((nickname, dict()) for nickname in nicknames)
        models = dict((nickname, dict()) for nickname in nicknames)
        for nickname, S in series.items():
            for dest, string, reporter, name in \
                    [(calibmaps[nickname], s['calib'], self.calibratereporter, names.nickname2calibrate_topic(nickname)), \
                    (models[nickname], s['model'], self.trainreporter, names.nickname2train_topic(nickname)),
                ]:
                for s in S:
                    if string not in dest:
                        start, end = [float(_) for _ in string.split("_")] ### NOTE: could be fragile!
                        dest[string] = reporter._read(reporter.path(name, start, end))

        # evaluated quivers
        logger.info('retrieving quivers within [%.3f, %.3f)'%(self.start, self.end))
        quivers = dict((nickname, features.Quiver()) for nickname in nicknames)
        for nickname in nicknames:
            for _, _, q in self.evaluatereporter.glob(names.nickname2evaluate_topic(nickname), self.start, self.end):
                q.filter(segs=[[self.start, self.end]])
                quivers[nickname].append_quiver(q)

        ### determine state
        logger.info('computing state and summary')
        logger.warn('state and summary currently set to defaults independent of data!')
        state, summary = self.state(nicknames, series, self.zoom_start, self.zoom_end, self.t0)

        ### generate a table summarizing the basic numbers for each nickname
        logger.info('computing basic statistics')
        cols = [
            'nickname',
            '% time with p(glitch|aux)>='+'%.3f within [%.3f, %.3f)'%(pglitch_thr, self.zoom_start, self.zoom_end),
            'number of glitch samples within [%.3f, %.3f)'%(self.start, self.end),
            'number of clean samples within [%.3f, %.3f)'%(self.start, self.end),
            'number of models within [%.3f, %.3f)'%(self.start, self.end),
            'number of calibration maps within [%.3f, %.3f)'%(self.start, self.end),
        ]
        data = []
        for nickname in nicknames:
            # do some quick data manipulations
            gch, cln = quivers[nickname].vectors2classes()
            for key in series[nickname][0]['data'].keys():
                if 'PGLITCH' in key:
                    break
            times = np.concatenate(tuple(s['data']['times'] for s in series[nickname]))
            pglitch = np.concatenate(tuple(s['data'][key] for s in series[nickname]))
            pglitch = pglitch[(self.zoom_start<=times)*(times<self.zoom_end)]
            # add a row to the table
            data.append([
                nickname,
                100.*np.sum(pglitch>=pglitch_thr)/len(pglitch),
                len(gch),
                len(cln),
                len(models[nickname]),
                len(calibmaps[nickname]),
            ])
        tables.append(dqr_json.format_table(cols, data))

        ### make plots
        # timeseries
        logger.info('generating timeseries plots')
        alttext = 'timeseries'

        gch_gps = []
        if annotate_gch_gps:
            for q in quivers.values():
                gch_gps += list(q.vectors2classes()[0].gpss)
            gch_gps = set(gch_gps)
        fig = plots.dqr_timeseries(nicknames, series, self.zoom_start, self.zoom_end, self.t0, gch_gps=gch_gps)
        fig.suptitle('%s Timeseries%s'%(self.instrument, title_graceid), fontsize=self._title_fontsize)
        path = self.path(graceid, name=alttext, output_dir=output_dir, suffix=figtype)
        logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)

        uploads.append((alttext, path))
        figures.append(dqr_json.format_figure(os.path.basename(path), alttext=alttext))

        # calbiration
        logger.info('generating calibration coverage plots')
        alttext = 'calibration_coverage'

        fig = plots.dqr_calibration_coverage(nicknames, series, quivers)
        fig.suptitle('%s Calibration Coverage %swithin [%.3f, %.3f)'%(self.instrument, title_graceid, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(graceid, name=alttext, output_dir=output_dir, suffix=figtype)
        logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)

        uploads.append((alttext, path))
        figures.append(dqr_json.format_figure(os.path.basename(path), alttext=alttext))

        # calibration distributions
        logger.info('generating calibration distribution plots')
        alttext = 'calibration_distribs'

        fig = plots.dqr_calibration_distribs(nicknames, quivers, calibmaps)
        fig.suptitle('%s Calibration Distributions %swithin [%.3f, %.3f)'%(self.instrument, title_graceid, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(graceid, name=alttext, output_dir=output_dir, suffix=figtype)
        logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)

        uploads.append((alttext, path))
        figures.append(dqr_json.format_figure(os.path.basename(path), alttext=alttext))

        # ROC curve
        logger.info('generating ROC plots')
        alttext = 'roc'

        fig = plots.dqr_roc(nicknames, series, quivers, calibmaps)
        fig.suptitle('%s Receiver Operating Characteristics %swithin [%.3f, %.3f)'%(self.instrument, title_graceid, self.start, self.end), fontsize=self._title_fontsize)
        path = self.path(graceid, name=alttext, output_dir=output_dir, suffix=figtype)
        logger.info('saving: '+path)
        fig.savefig(path)
        plots.close(fig)

        uploads.append((alttext, path))
        figures.append(dqr_json.format_figure(os.path.basename(path), alttext=alttext))

        # feature importance
        logger.info('generating feature importance plots and tables')

        # instantiate classifiers to pick up some standard stuff...
        classifier_factory = factories.ClassifierFactory()
        klassifiers = dict((nickname, classifier_factory(nickname, rootdir=output_dir, **configparser.evalkwargs(**dict(self.config.items(nickname))))) for nickname in nicknames)

        # actually compute stuff
        figs, tabs = plots.dqr_featureimportance(nicknames, models, klassifiers, self.quiver_factory, self.zoom_start, self.zoom_end, self.t0)
        for nickname in nicknames:
            for key, model in models[nickname].items():
                fig = figs[nickname][key]
                tab = tabs[nickname][key]
                if (fig is None) and (tab is None):
                    logger.warn('feature importance not available for nickname='+nickname)

                if fig is not None:
                    alttext = nickname+'-'+self.instrument+'_featureimportance-'+model.hash
                    path = self.path(graceid, name=alttext, output_dir=output_dir, suffix=figtype)
                    fig.suptitle('%s %s Feature Importance %swithin [%.3f, %.3f)'%(self.instrument, nickname, title_graceid, self.start, self.end), fontsize=self._title_fontsize)
                    logger.info('saving: '+path)
                    fig.savefig(path)
                    plots.close(fig)

                    uploads.append((alttext, path))
                    figures.append(dqr_json.format_figure(os.path.basename(path), alttext=alttext))

                if tab is not None:
                    cols, data = tab
                    tmp = nickname+" "+model.hash+" %s" ### make sure the columns are clearly labeled...
                    cols = [tmp%_ for _ in cols]
                    tables.append(dqr_json.format_table(cols, data))

        ### write report
        reportpath = self.path(graceid, output_dir=output_dir, suffix='json')
        logger.info('writing DQR JSON report to: '+reportpath)
        with open(reportpath, 'w') as obj:
            json.dump(
                dqr_json.format_report(
                    state,
                    __process_name__,
                    __version__,
                    librarian=__author__,
                    summary=summary,
                    figures=figures,
                    tables=tables,
                    links=[dqr_json.format_link(*_) for _ in links],
                ),
                obj,
            )
        uploads.append(('iDQ report', reportpath))

        # upload everything to GraceDb
        if graceid is not None:
            logger.info('uploading figures, report to %s in %s'%(graceid, gracedb_url))

            if GraceDb is None:
                raise ImportError('could not import ligo.gracedb.rest.GraceDb')

            client = GraceDb(gracedb_url)
            for message, path in uploads: ### upload all figures
                logger.info('uploading %s'%path)
                client.writeLog(graceid, message, filename=path, tag_name=[__process_name__])

    def path(self, graceid, name='', output_dir='.', suffix='png'):
        return os.path.join(output_dir, dqr_logging.get_nickname(self.instrument+__process_name__+self.tag, graceid=graceid, tag=name)+'.'+suffix)
            
    @staticmethod
    def state(nicknames, series, start, end, t0):
        state = "human_input_needed" ### FIXME: make a better choice here! something based on data would be nice...
        summary = "WARNING: iDQ currently always returns \"human_input_needed\"!"

        # compute the quantile for pglitch at t0
        for nickname in nicknames:
            for key in series[nickname][0]['data'].keys():
                if 'PGLITCH' in key:
                    break 
            pglitch = []
            for s in series[nickname]:
                pglitch.append(s['data'][key])
                times = s['data']['times']
                if (times[0]<=t0) and (t0<times[-1]+s['deltaT']):
                    p0 = np.interp(t0, times, s['data'][key])
                    break
            else:
                segs = utils.fix_segments([(s['data']['times'][0], s['data']['times'][-1]+s['deltaT']) for s in series[nickname]])
                raise ValueError('t0 (%.3f) not containd in any of the series discovered: %s'%(t0, '; '.join('[%.3f, %.3f)'%(s, e) for s, e in segs)))

            pglitch = np.concatenate(tuple(pglitch))
            summary += '<br>%s p(glitch|aux)=%.3f at %.3f. %.3f%s of the time within [%.3f, %.3f) has a higher %s p(glitch|aux)'%\
                (nickname, p0, t0, 100.*np.sum(pglitch>=p0)/len(pglitch), '%', start, end, nickname)

        return state, summary
